from distutils.log import WARN
from logging import WARNING
from unittest import TestCase


class TestLogging(TestCase):
    def test_import_logging(self):
        from splight_lib import logging
        logger = logging.getLogger()
        logger.debug("DEBUG MESSAGE")
        logger.info("INFO MESSAGE")
        logger.warning("WARNING MESSAGE")
        logger.critical("CRITICAL MESSAGE")
        try:
            raise Exception("EXCEPTION MESSAGE")
        except Exception as e:
            logger.exception(e)

    def test_logger_message_present(self):
        from splight_lib import logging
        logger = logging.getLogger()
        message = "WARNING MESSAGE"
        with self.assertLogs() as captured:
            logger.warning(message)
        self.assertEqual(len(captured.records), 1)
        self.assertEqual(captured.records[0].getMessage(), message)
        self.assertEqual(captured.records[0].levelno, WARNING)



from splight_models.communication.context import *
from splight_models.communication.events import *


__all__ = [
    "CommunicationChannelData",
    "CommunicationClientStatus",
    "CommunicationContext",
    "CommunicationEvent",
    "EventNames",
    "EventActions",
    # TODO remove this
    "MappingCreateEvent",
    "MappingDeleteEvent",
]


import uuid
from pydantic import BaseModel
from typing import List, Type
from splight_abstract import (
    AbstractDatabaseClient,
    validate_instance_type,
    validate_resource_type
)
from splight_lib import logging
from typing import Dict
from splight_models import *
from collections import defaultdict
from tempfile import TemporaryFile

logger = logging.getLogger()


class FakeDatabaseClient(AbstractDatabaseClient):
    database: Dict[Type, List[BaseModel]] = defaultdict(list)
    valid_classes = [
        Mapping,
        Asset,
        Attribute,
        File,
        Tag,
        Namespace,
        Component,
        Notification,
        Graph,
        Node,
        Edge,
        BlockchainContract,
        BlockchainContractSubscription,
        Component,
        ComponentObject,
        ComponentCommand,
        Query,
    ]

    def _create(self, instance: BaseModel) -> BaseModel:
        instance.id = str(uuid.uuid4())
        self.database[type(instance)].append(instance)
        return instance

    @validate_instance_type
    def save(self, instance: BaseModel) -> BaseModel:
        logger.debug(f"[FAKED] Executing save with {instance}")

        for i, item in enumerate(self.database[type(instance)]):
            if item.id == instance.id:
                self.database[type(instance)][i] = instance
                return instance
        return self._create(instance)

    @validate_resource_type
    def _get(self, resource_type: Type,
             first: bool = False,
             limit_: int = -1,
             skip_: int = 0,
             **kwargs) -> List[BaseModel]:
        logger.debug(f"[FAKED] Executing get with {resource_type}")
        queryset = self.database[resource_type]
        kwargs = self._validated_kwargs(resource_type, **kwargs)
        queryset = self._filter(queryset, **kwargs)
        if limit_ != -1:
            queryset = queryset[skip_:skip_ + limit_]

        if first:
            return queryset[0] if queryset else None

        return queryset

    @validate_resource_type
    def count(self, resource_type: Type, **kwargs) -> int:
        return len(self._get(resource_type, **kwargs))

    @validate_resource_type
    def delete(self, resource_type: Type, id: str) -> None:
        logger.debug(f"[FAKED] Executing delete with {resource_type} {id}")

        queryset = self.database.get(resource_type, [])
        for i, item in enumerate(queryset):
            if item.id == id:
                del queryset[i]
                return

    @validate_instance_type
    def download(self, instance: BaseModel) -> None:
        f = TemporaryFile("w+")
        data = instance.dict()
        json.dump(data, f, indent=4)
        f.seek(0)
        return f



class AuthenticationError(Exception):
    def __init__(self, url: str):
        self._msg = "An error ocurred during authentication"

    def __str__(self) -> str:
        return self._msg


class SignatureVerificationError(Exception):
    def __init__(self, message, sig_header, http_body=None):
        super(SignatureVerificationError, self).__init__(message)
        self.sig_header = sig_header
        self._message = message
        self.http_body = http_body

    def __repr__(self):
        return f"{self.message} {self.http_body}"

    def __str__(self):
        return self._message



from .classmap import CLASSMAP
from retry import retry
from splight_abstract.remote import AbstractRemoteClient
from splight_abstract.database import AbstractDatabaseClient
from splight_models import File
from remote_splight_lib.settings import settings
from remote_splight_lib.exceptions import InvalidModel
from remote_splight_lib.auth import SplightAuthToken
import json
from typing import Dict, List, Type

from furl import furl
from pydantic import BaseModel
from requests import Session
from requests.exceptions import (
    ConnectionError,
    Timeout
)
from tempfile import TemporaryFile


REQUEST_EXCEPTIONS = (ConnectionError, Timeout)


class DatabaseClient(AbstractDatabaseClient, AbstractRemoteClient):
    """Splight API Database Client.
    Responsible for interacting with database resources using HTTP requests
    to the Splight API.
    """

    def __init__(self, namespace: str = "default"):
        super(DatabaseClient, self).__init__(namespace=namespace)
        self._base_url = furl(settings.SPLIGHT_PLATFORM_API_HOST)
        token = SplightAuthToken(
            access_key=settings.SPLIGHT_ACCESS_ID,
            secret_key=settings.SPLIGHT_SECRET_KEY,
        )
        self._session = Session()
        self._session.headers.update(token.header)

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=1)
    def save(self, instance: BaseModel) -> BaseModel:
        """Creates or updates a new resource depending on the model if
        it contains the id or not.

        Parameters
        ----------
        instance : BaseModel
            The instance of the model to be created or updated

        Returns
        -------
        BaseModel with the created or updated resource.

        Raises
        ------
        InvalidModel thrown when the model name is not correct.
        """
        constructor = type(instance)
        model_data = self._get_model_data(constructor)

        path = model_data["path"]
        if instance.id:
            output = self._update(path, instance.id, instance)
        else:
            output = self._create(path, instance)
        return constructor.parse_obj(output)

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=1)
    def delete(self, resource_type: Type, id: str):
        """Deletes a resource from the database

        Parameters
        ----------
        resource_type : Type
            The resource type to be deleted
        id : str
            The resource's id.

        Raises
        ------
        InvalidModel thrown when the model name is not correct.
        """
        model_data = self._get_model_data(resource_type)
        path = model_data["path"]
        url = self._base_url / f"{path}/{id}/"
        response = self._session.delete(url)
        response.raise_for_status()

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=1)
    def _get(
        self,
        resource_type: Type,
        first: bool = False,
        limit_: int = -1,
        skip_: int = 0,
        deleted: bool = False,
        **kwargs,
    ) -> List[BaseModel]:
        model_data = self._get_model_data(resource_type)
        path = model_data["path"]
        response = self._list(
            path,
            limit_=limit_,
            skip_=skip_,
            deleted=deleted,
            **kwargs,
        )
        parsed = [
            resource_type.parse_obj(resource)
            for resource in response["results"]
        ]
        return parsed

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=1)
    def count(self, resource_type: Type, **kwargs) -> int:
        """Returns the number of resources in the database for a given model

        Parameters
        ----------
        resource_type : str
            The model name

        Returns
        -------
        int

        Raises
        ------
        InvalidModel thrown when the model name is not correct.
        """
        model_data = self._get_model_data(resource_type)
        path = model_data["path"]
        kwargs["page"] = 1  # Always start from the first page
        response = self._list(path, **kwargs)
        return response["count"]

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=1)
    def download(self, instance: BaseModel, **kwargs) -> TemporaryFile:
        """Returns the number of resources in the database for a given model

        Parameters
        ----------
        instance : BaseModel
            The instance of the model to be downloaded

        Returns
        -------
        TemporaryFile
            the file object

        Raises
        ------
        InvalidModel thrown when the model name is not correct.
        """
        constructor = type(instance)
        model_data = self._get_model_data(constructor)
        path = model_data["path"]
        url = self._base_url / f"{path}/{instance.id}/download"
        response = self._session.get(url)
        response.raise_for_status()
        f = TemporaryFile("wb+")
        f.write(response.content)
        f.seek(0)
        return f

    def _pages(self, path: str, **kwargs):
        next_page = kwargs["page"]
        while next_page:
            response = self._list(path, **kwargs)
            yield response
            next_page = (
                response["next"].split("page=")[1]
                if response["next"]
                else None
            )
            kwargs["page"] = next_page

    def _list(self, path: str, **kwargs):
        url = self._base_url / f"{path}/"
        params = self._parse_params(**kwargs)
        response = self._session.get(url, params=params)
        response.raise_for_status()
        return response.json()

    @staticmethod
    def _get_model_data(constructor: Type):
        model_data = CLASSMAP.get(constructor)
        if not model_data:
            raise InvalidModel(constructor.schema()["title"])
        return model_data

    def _create(self, path: str, instance: BaseModel) -> Dict:
        url = self._base_url / f"{path}/"
        data = json.loads(instance.json(exclude_none=True))
        if isinstance(instance, File):
            with open(instance.file, 'rb') as f:
                file = {"file": f}
                response = self._session.post(url, data=data, files=file)
        else:
            response = self._session.post(url, json=data)

        response.raise_for_status()
        return response.json()

    def _update(self, path: str, resource_id: str, data: BaseModel) -> Dict:
        url = self._base_url / f"{path}/{resource_id}/"
        response = self._session.put(
            url, json=json.loads(data.json(exclude_none=True))
        )
        response.raise_for_status()
        return response.json()



from datetime import timedelta, timezone
from io import StringIO
from tempfile import NamedTemporaryFile
from typing import Dict, List, Type, Union

import pandas as pd
import json
from furl import furl
from pydantic import BaseModel
from requests import Session

from remote_splight_lib.auth import SplightAuthToken
from remote_splight_lib.settings import settings
from splight_abstract.remote import AbstractRemoteClient
from splight_abstract.datalake import AbstractDatalakeClient, validate_resource_type
from splight_models import Query
from retry import retry

from requests.exceptions import (
    ConnectionError,
    Timeout,
    HTTPError,
)
REQUEST_EXCEPTIONS = (ConnectionError, Timeout, HTTPError)


class DatalakeClient(AbstractDatalakeClient, AbstractRemoteClient):

    _PREFIX = "datalake"

    def __init__(self, namespace: str = "default"):
        super(DatalakeClient, self).__init__(namespace=namespace)
        self._base_url = furl(settings.SPLIGHT_PLATFORM_API_HOST)
        token = SplightAuthToken(
            access_key=settings.SPLIGHT_ACCESS_ID,
            secret_key=settings.SPLIGHT_SECRET_KEY,
        )
        self._session = Session()
        self._session.headers.update(token.header)

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=2, jitter=1)
    def save(
        self,
        instances: List[BaseModel],
        collection: str = "default",
    ) -> List[BaseModel]:
        # POST /datalake/save/
        url = self._base_url / f"{self._PREFIX}/save/"
        data = [json.loads(model.json()) for model in instances]
        response = self._session.post(
            url, params={"source": collection}, json=data
        )
        response.raise_for_status()
        return instances

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=2, jitter=1)
    def raw_save(
        self,
        instances: List[Dict],
        collection: str = "default",
    ) -> List[Dict]:
        # POST /datalake/save/
        url = self._base_url / f"{self._PREFIX}/save/"
        response = self._session.post(
            url, params={"source": collection}, json=instances
        )
        response.raise_for_status()
        return instances

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=2, jitter=1)
    def _raw_get(self,
                 collection: str = 'default',
                 limit_: int = 50,
                 skip_: int = 0,
                 sort: Union[List, str] = ['timestamp__desc'],
                 group_id: Union[List, str] = [],
                 group_fields: Union[List, str] = [],
                 tzinfo: timezone = timezone(timedelta()),
                 **kwargs) -> List[BaseModel]:
        # /datalake/data/
        url = self._base_url / f"{self._PREFIX}/data/"
        kwargs.update(
            {
                "source": collection,
                "limit_": limit_,
                "skip_": skip_,
                "sort": sort,
                # "tzinfo": tzinfo
            }
        )
        if group_id:
            kwargs.update({"group_id": group_id})
        if group_fields:
            kwargs.update({"group_fields": group_fields})
        params = self._parse_params(**kwargs)
        response = self._session.get(url, params=params)
        response.raise_for_status()
        return response.json()["results"]

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=2, jitter=1)
    @validate_resource_type
    def _get(
        self,
        resource_type: Type,
        collection: str = "default",
        limit_: int = 50,
        skip_: int = 0,
        sort: Union[List, str] = ["timestamp__desc"],
        group_id: Union[List, str] = [],
        group_fields: Union[List, str] = [],
        tzinfo: timezone = timezone(timedelta()),
        **kwargs,
    ) -> List[BaseModel]:
        # /datalake/data/
        url = self._base_url / f"{self._PREFIX}/data/"
        valid_kwargs = self._validated_kwargs(resource_type, **kwargs)
        valid_kwargs.update(
            {
                "source": collection,
                "limit_": limit_,
                "skip_": skip_,
                "sort": sort,
                # "tzinfo": tzinfo
            }
        )
        if group_id:
            valid_kwargs.update({"group_id": group_id})
        if group_fields:
            valid_kwargs.update({"group_fields": group_fields})
        params = self._parse_params(**valid_kwargs)
        response = self._session.get(url, params=params)
        response.raise_for_status()
        output = [
            resource_type.parse_obj(item)
            for item in response.json()["results"]
        ]
        return output

    # TODO: Add add_fields, project and renaming
    def get_output(self, query: Query) -> List[Dict]:
        return self.raw_get(
            collection=query.collection,
            limit_=query.limit,
            skip_=query.skip,
            sort=query.sort,
            group_id=query.group_id,
            group_fields=query.group_fields,
            tzinfo=timezone(timedelta(hours=query.timezone_offset)),
            **query.filters
        )

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=2, jitter=1)
    def get_dataframe(
        self,
        freq: str = None,
        collection: str = "default",
        **kwargs
    ) -> pd.DataFrame:
        # GET /datalake/dumpdata/?source=collection
        url = self._base_url / f"{self._PREFIX}/dumpdata/"
        kwargs.update({"source": collection})
        kwargs.update({"freq": freq})
        params = self._parse_params(**kwargs)
        response = self._session.get(url, params=params)
        response.raise_for_status()

        df: pd.DataFrame = pd.DataFrame(pd.read_csv(StringIO(response.text)))
        if df.empty:
            return df
        df["timestamp"] = pd.to_datetime(df["timestamp"])
        df.set_index("timestamp", inplace=True)
        return df

    def get_dataset(self, queries: List[Dict]) -> pd.DataFrame:
        dfs = [self.get_dataframe(**query) for query in queries]
        df = pd.concat(dfs, axis=1)
        return df

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=2, jitter=1)
    def save_dataframe(
        self, dataframe: pd.DataFrame, collection: str = 'default'
    ) -> None:
        # POST /datalake/loaddata/
        url = self._base_url / f"{self._PREFIX}/loaddata/"

        tmp_file = NamedTemporaryFile("w")
        with open(tmp_file.name, "wb") as fid:
            dataframe.to_csv(fid)

        response = self._session.post(
            url,
            data={"source": collection},
            files={"file": open(tmp_file.name)},
        )
        response.raise_for_status()

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=2, jitter=1)
    def list_collection_names(self) -> List[str]:
        # GET /datalake/source
        url = self._base_url / f"{self._PREFIX}/source/"
        response = self._session.get(url)
        response.raise_for_status()
        sources = [source["source"] for source in response.json()["results"]]
        return sources

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=2, jitter=1)
    def get_unique_keys(self, collection: str) -> List[str]:
        # GET /datalake/key/?source=collection
        url = self._base_url / f"{self._PREFIX}/key/"
        response = self._session.get(url, params={"source": collection})
        response.raise_for_status()
        keys = list({key["key"] for key in response.json()["results"]})
        return keys

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=2, jitter=1)
    def get_values_for_key(self, collection: str, key: str) -> List[str]:
        # GET /datalake/value/?source=collection&key=key
        url = self._base_url / f"{self._PREFIX}/value/"
        params = {"source": collection, "key": key}
        response = self._session.get(url, params=params)
        response.raise_for_status()
        values = [value["value"] for value in response.json()["results"]]
        return values

    # Subject to incompatibility by implementation
    @retry(REQUEST_EXCEPTIONS, tries=3, delay=2, jitter=1)
    def raw_aggregate(
        self, collection: str, pipeline: List[Dict]
    ) -> List[Dict]:
        # POST /datalake/aggregate/?source=collection
        url = self._base_url / f"{self._PREFIX}/aggregate/"
        params = {"source": collection}
        data = {"pipeline": pipeline}
        response = self._session.post(url, params=params, data=data)
        response.raise_for_status()
        return response.json()

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=2, jitter=1)
    def get_db_size_gb(self) -> float:
        # GET /datalake/db-size/
        url = self._base_url / f"{self._PREFIX}/db-size/"
        response = self._session.get(url)
        response.raise_for_status()
        return response.json()

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=2, jitter=1)
    def create_index(self, collection: str, index: list) -> None:
        # POST /datalake/index/
        url = self._base_url / f"{self._PREFIX}/index/"
        data = {"source": collection, "index": index}
        response = self._session.post(url, json=data)
        response.raise_for_status()



from unittest import TestCase
from splight_abstract.client.hooks import HooksMixin


class ForTest(HooksMixin):
    def test(self, a, b, **kwargs):
        return a, b, kwargs


class TestVariable(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.instance = ForTest()

    @staticmethod
    def a_plus(*args, **kwargs):
        kwargs['a'] += 1
        return args, kwargs

    @staticmethod
    def b_minus(*args, **kwargs):
        kwargs['b'] -= 1
        return args, kwargs

    @staticmethod
    def join_return_values(result):
        a, b, kwargs = result
        return_value = {
            "a": a,
            "b": b,
        }
        return_value.update(**kwargs)
        return return_value

    def test_post_hook(self):
        self.instance.add_post_hook('test', self.join_return_values)
        return_value = self.instance.test(1, 2, c=3)
        self.assertEqual(return_value, {"a": 1, "b":2 , "c":3})

    def test_1_hook(self):
        self.instance.add_pre_hook('test', self.a_plus)

        a, b, kwargs = self.instance.test(1, 2)
        self.assertEqual(a, 2)
        self.assertEqual(b, 2)
        self.assertEqual(kwargs, {})

    def test_2_hook(self):
        self.instance.add_pre_hook('test', self.a_plus)
        self.instance.add_pre_hook('test', self.b_minus)

        a, b, kwargs = self.instance.test(1, 2)
        self.assertEqual(a, 2)
        self.assertEqual(b, 1)
        self.assertEqual(kwargs, {})

    def test_kwargs(self):
        self.instance.add_pre_hook('test', self.a_plus)

        a, b, kwargs = self.instance.test(1, 2, p=3, q=4)
        self.assertEqual(a, 2)
        self.assertEqual(kwargs, {'p': 3, 'q': 4})



import logging
import time
import os


def getLogger(name=None):
    level = int(os.getenv('LOG_LEVEL', logging.DEBUG))
    log_file = os.getenv('LOG_FILE', None)
    fmt = '%(levelname)s | %(asctime)s | %(filename)s:%(lineno)d | %(message)s'
    config = {
        'level': level,
        'format': fmt,
    }
    if log_file:
        config['filename'] = log_file
        config['filemode'] = 'a'

    logging.basicConfig(**config)
    logging.Formatter.converter = time.gmtime

    return logging.getLogger(name)




from enum import Enum
from typing import Dict, Optional
from pydantic import Field
from splight_models.base import SplightBaseModel
from splight_models.mapping import Mapping
from splight_models.user import User
from datetime import datetime, timezone


class EventActions(str, Enum):
    CREATE = "CREATE"
    UPDATE = "UPDATE"
    DELETE = "DELETE"
    TRIGGER = "TRIGGER"
    READ = "READ"


class EventNames(str, Enum):
    # TODO make this use EventActions.
    COMPONENT_COMMAND_TRIGGER = 'componentcommand-trigger'
    COMPONENT_COMMAND_CREATE = 'componentcommand-create'
    COMPONENT_COMMAND_UPDATE = 'componentcommand-update'
    # TODO add Asset Attribute and all shared objects
    # TODO delete these
    MAPPING_CREATE = "mapping-create"
    MAPPING_DELETE = "mapping-delete"

class CommunicationEvent(SplightBaseModel):
    event_name: str
    id: Optional[str] = None
    instance_id: Optional[str] = None
    socket_id: Optional[str] = None
    timestamp: str = Field(default_factory=lambda: datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%fZ"))  # pusher cannot json serialize datetime objects
    display_text: Optional[str] = None
    user: Optional[User] = None
    data: Dict


# TODO remove all this
class MappingCreateEvent(CommunicationEvent):
    event_name: str = Field(EventNames.MAPPING_CREATE, const=True)
    data: Mapping


class MappingDeleteEvent(CommunicationEvent):
    event_name: str = Field(EventNames.MAPPING_DELETE, const=True)
    data: Mapping



import inspect
from enum import Enum
from typing import Callable, List, Dict
from functools import wraps


class HooksStage(str, Enum):
    BEFORE = "before"
    AFTER = "after"


class HooksMixin:
    _original_signature = {}

    def add_pre_hook(self, function, hook):
        setattr(self, function, self._add_hook(getattr(self, function), hook, HooksStage.BEFORE))

    def add_post_hook(self, function, hook):
        setattr(self, function, self._add_hook(getattr(self, function), hook, HooksStage.AFTER))

    def _add_hook(self, func, hook, stage: HooksStage = HooksStage.BEFORE):
        if func.__name__ not in self._original_signature:
            self._original_signature[func.__name__] = [
                param
                for param in inspect.getfullargspec(func).args
                if param not in ['self', 'cls', 'args', 'kwargs']
            ]

        @wraps(func)
        def pre_hooked_func(*args, **kwargs):
            kwargs = self._args_to_kwargs(func, *args, **kwargs)
            args, kwargs = hook(**kwargs)
            return func(*args, **kwargs)

        @wraps(func)
        def post_hooked_func(*args, **kwargs):
            kwargs = self._args_to_kwargs(func, *args, **kwargs)
            result = func(**kwargs)
            return hook(result)
        return pre_hooked_func if stage == HooksStage.BEFORE else post_hooked_func

    def _args_to_kwargs(self, func: Callable, *args, **kwargs) -> Dict:
        if not args:
            return kwargs
        func_params: List = self._original_signature[func.__name__]
        new_kwargs: Dict = dict(zip(func_params, args))
        new_kwargs.update(kwargs)

        return new_kwargs



from uuid import uuid4
from asyncio.log import logger
from typing import Any, Dict

from splight_abstract import AbstractAuthClient, QuerySet
from splight_models import User


class FakeCredentials:
    def count(self, *args, **kwargs):
        return self.list()[0]["count"]

    def get(self, *args, **kwargs) -> QuerySet:
        return QuerySet(self, *args, **kwargs)

    def _get(self, *args, **kwargs):
        return self.list()[0]["results"]

    def list(self):
        return (
            {
                "count": 3,
                "next": None,
                "previous": None,
                "results": [
                    {
                        "access_id": "6f5c1652-c52d-4aea-bd9c-5bf72d63c5b6",
                        "created_date": "2022-07-05T18:02:32.692815Z",
                        "last_used": "2022-07-05T18:02:32.692824Z",
                        "user": 1,
                    },
                    {
                        "access_id": "01c5bc01-d0c6-43b1-91d2-698f80545c82",
                        "created_date": "2022-07-05T18:02:35.630840Z",
                        "last_used": "2022-07-05T18:02:35.630856Z",
                        "user": 1,
                    },
                    {
                        "access_id": "13e9294c-4810-4a8d-8452-61961c7e8ecd",
                        "created_date": "2022-07-05T18:02:36.916698Z",
                        "last_used": "2022-07-05T18:02:36.916708Z",
                        "user": 1,
                    },
                ],
            },
            200,
        )

    def create(self, data: Dict[str, Any]):
        secret_key = (
            "addffg5b024dbc7f6b6760012ea493e738e350fad0b19e02405342f6666fd513"
        )
        return (
            {
                "access_id": "57e45010-ad7a-4dce-b759-98ec4f77a079",
                "created_date": "2022-07-05T19:29:16.061270Z",
                "last_used": "2022-07-05T19:29:16.061282Z",
                "user": 1,
                "secret_key": secret_key,
            },
            201,
        )

    def destroy(self, resource_id: str):
        return ({}, 204)


class FakeDeployment:
    def list(self):
        return (
            {
                "count": 3,
                "next": None,
                "previous": None,
                "results": [
                    {
                        "access_id": "6f5c1652-c52d-4aea-bd9c-5bf72d63c5b6",
                        "created_date": "2022-07-05T18:02:32.692815Z",
                        "last_used": "2022-07-05T18:02:32.692824Z",
                    },
                    {
                        "access_id": "01c5bc01-d0c6-43b1-91d2-698f80545c82",
                        "created_date": "2022-07-05T18:02:35.630840Z",
                        "last_used": "2022-07-05T18:02:35.630856Z",
                    },
                    {
                        "access_id": "13e9294c-4810-4a8d-8452-61961c7e8ecd",
                        "created_date": "2022-07-05T18:02:36.916698Z",
                        "last_used": "2022-07-05T18:02:36.916708Z",
                    },
                ],
            },
            200,
        )

    def create(self, data: Dict[str, Any]):
        secret_key = (
            "addffg5b024dbc7f6b6760012ea493e738e350fad0b19e02405342f6666fd513"
        )
        return (
            {
                "access_id": "57e45010-ad7a-4dce-b759-98ec4f77a079",
                "created_date": "2022-07-05T19:29:16.061270Z",
                "last_used": "2022-07-05T19:29:16.061282Z",
                "secret_key": secret_key,
            },
            201,
        )

    def destroy(self, resource_id: str):
        return ({}, 204)


class FakeProfile:
    def count(self, *args, **kwargs):
        return self.list()[0]["count"]

    def get(self, *args, **kwargs) -> QuerySet:
        return QuerySet(self, *args, **kwargs)

    def _get(self, *args, **kwargs):
        return self.list()[0]["results"]

    def list(self):
        return (
            {
                "organization_id": "org_gxUvmxKLxSZPtHH8",
                "original_organization_id": "org_gxUvmxKLxSZPtHH3",
                "user_id": "auth0|628f9e023211690069ccc193",
                "username": "fake.user",
                "name": "fake.user@splight-ae.com",
                "email": "fake.user@splight-ae.com",
                "picture": "https://s.gravatar.com/avatar/a7b54dab1cd2b75ec1b1a8f6ed31efdc?s=480&r=pg&d=https%3A%2F%2Fcdn.auth0.com%2Favatars%2Fma.png",
                "permissions": ["admin", "splightadmin"],
                "user_metadata": {
                    "language": "en",
                    "theme": "light",
                    "picture_color": "light-green",
                },
                "app_metadata": {"is_manager": True},
                "role": "splightadmin",
            },
            200,
        )

    def update_profile(self, data: Dict[str, Any]):
        return self.list()


class FakeOrganization:

    def count(self, *args, **kwargs):
        return self.list()[0]["count"]

    def get(self, *args, **kwargs) -> QuerySet:
        return QuerySet(self, *args, **kwargs)

    def _get(self, *args, **kwargs):
        return self.list()[0]

    def list(self):
        return (
            {
                "id": "org_fGMkta1hcU4NWkp1",
                "name": "splight",
                "display_name": "Splight",
                "metadata": {
                    "billing_id": "123",
                    "blockchain_id": (
                        "0xcCd6872568009c5669fc699098a4373171C905B6"
                    ),
                    "manager_email": "fake.user@splight-ae.com"
                },
            },
            200,
        )

    def update_metadata(self, data: Dict[str, Any]):
        return self.list()


class FakeRoles:

    def count(self, *args, **kwargs):
        return self.list()[0]["count"]

    def get(self, *args, **kwargs) -> QuerySet:
        return QuerySet(self, *args, **kwargs)

    def _get(self, *args, **kwargs):
        return self.list()[0]["results"]

    def list(self):
        return (
            {
                "count": 4,
                "next": None,
                "previous": None,
                "results": [
                    {
                        "id": "rol_xYsOLG4Kni9I9EJF",
                        "name": "admin",
                        "description": None,
                        "permissions": [],
                    },
                    {
                        "id": "rol_GZEhX5vu4WbVWeM5",
                        "name": "editor",
                        "description": None,
                        "permissions": [],
                    },
                    {
                        "id": "rol_8O7rXRkdtRklGW28",
                        "name": "splightadmin",
                        "description": None,
                        "permissions": [],
                    },
                    {
                        "id": "rol_eKzFB2R4lqYxeF2K",
                        "name": "viewer",
                        "description": None,
                        "permissions": [],
                    },
                ],
            },
            200,
        )

    def retrieve(self, resource_id: str):
        return self.list()[0]["results"][0]


class FakeUsers:
    FAKE_USER_0 = {
        "user_id": "google-oauth2|109709194253657961436",
        "email": "maximiliano.pezzotta@splight-ae.com",
        "picture": (
            "https://lh3.googleusercontent.com/a/AATXAJx7mtKTKLwpBNLy0TOQJ0jK"
            "5Y2LMcbEzV4xQBNF=s96-c"
        ),
        "name": "Maximiliano Pezzotta",
        "app_metadata": {"is_manager": False},
        "user_metadata": {"theme": "dark", "theme": "dark"},
        "blocked": False,
        "roles": None,
        "role": None,
    }
    FAKE_USER_1 = {
        "user_id": "auth0|61571fb933f692007043a0aa",
        "email": "example@gmail.com",
        "picture": (
            "https://s.gravatar.com/avatar/e820bb4aba5ad74c5a6ff1aca16641f6?s"
            "=480&r=pg&d=https%3A%2F%2Fcdn.auth0.com%2Favatars%2Fex.png"
        ),
        "name": "example@gmail.com",
        "app_metadata": {"is_manager": False},
        "user_metadata": {},
        "blocked": False,
        "roles": None,
        "role": None,
    }
    FAKE_USER_2 = {
        "user_id": "google-oauth2|111755134701574065193",
        "email": "estrella.cholod@splight-ae.com",
        "picture": (
            "https://lh3.googleusercontent.com/a/AATXAJwi2Z-TlTLz96K6IPClNWef"
            "HNkAdu31RKAzSzxl=s96-c"
        ),
        "name": "Estrella Cholod",
        "app_metadata": {"is_manager": True},
        "user_metadata": {},
        "blocked": False,
        "roles": None,
        "role": None,
    }

    def count(self, *args, **kwargs):
        return self.list()[0]["count"]

    def get(self, *args, **kwargs) -> QuerySet:
        return QuerySet(self, *args, **kwargs)

    def _get(self, *args, **kwargs):
        return self.list()[0]["results"]

    def list(self):
        return (
            {
                "count": 3,
                "next": None,
                "previous": None,
                "results": [
                    self.FAKE_USER_0,
                    self.FAKE_USER_1,
                    self.FAKE_USER_2,
                ],
            },
            200,
        )

    def retrieve(self, resource_id: str):
        return (self.FAKE_USER_0, 200)

    def users_update(self, resource_id: str, data: Dict[str, Any]):
        return (self.FAKE_USER_1, 200)


class FakeInvitations:

    def count(self, *args, **kwargs):
        return self.list()[0]["count"]

    def get(self, *args, **kwargs) -> QuerySet:
        return QuerySet(self, *args, **kwargs)

    def _get(self, *args, **kwargs):
        return self.list()[0]["results"]

    def list(self):
        return (
            {
                "count": 2,
                "next": None,
                "previous": None,
                "results": [
                    {
                        "id": "uinv_4SfR3ucBxOj3wpqG",
                        "client_id": "1IsJjtPUv1DgHqsfZX4YsEjEStRRu1SG",
                        "inviter": {"name": "matias.silva@splight-ae.com"},
                        "invitee": {
                            "email": "matias.silva+1234@splight-ae.com"
                        },
                        "app_metadata": {},
                        "user_metadata": {},
                        "invitation_url": "https://integrationapp.splight-ae.com/auth-redirect?orgInvitation=true&invitation=kyWlQxQdZrX4RegzIFKhqHBwUcCGEZAj&organization=org_gxUvmxKLxSZPtHH8&organization_name=splight",
                        "ticket_id": "kyWlQxQdZrX4RegzIFKhqHBwUcCGEZAj",
                        "created_at": "2022-08-04T19:34:59.300Z",
                        "expires_at": "2022-08-11T19:34:59.300Z",
                        "organization_id": "org_gxUvmxKLxSZPtHH8",
                        "roles": ["rol_xYsOLG4Kni9I9EJF"],
                    },
                    {
                        "id": "uinv_0mKqsA8TqAZi2hCU",
                        "client_id": "1IsJjtPUv1DgHqsfZX4YsEjEStRRu1SG",
                        "inviter": {"name": "matias.silva@splight-ae.com"},
                        "invitee": {"email": "matias.silva+5@splight-ae.com"},
                        "app_metadata": {},
                        "user_metadata": {},
                        "invitation_url": "https://integrationapp.splight-ae.com/auth-redirect?orgInvitation=true&invitation=oq54fST3OQt3rxX7DI5wIDQbfudzKD9J&organization=org_gxUvmxKLxSZPtHH8&organization_name=splight",
                        "ticket_id": "oq54fST3OQt3rxX7DI5wIDQbfudzKD9J",
                        "created_at": "2022-08-04T17:47:48.639Z",
                        "expires_at": "2022-08-11T17:47:48.639Z",
                        "organization_id": "org_gxUvmxKLxSZPtHH8",
                        "roles": ["rol_xYsOLG4Kni9I9EJF"],
                    },
                ],
            },
            200,
        )

    def invitations_create(self, data: Dict[str, Any]):
        return (
            {
                "id": "asd",
                "client_id": "asd",
                "inviter": {"name": "fake.user@splight-ae.com"},
                "invitee": {"email": "fake.user+15@splight-ae.com"},
                "app_metadata": {},
                "user_metadata": {},
                "invitation_url": "https://integrationapp.splight-ae.com/",
                "ticket_id": "5qGQxSZJs9sQGDzmG1xAWhzTo0eabKVA",
                "created_at": "2022-07-05T19:37:48.367Z",
                "expires_at": "2022-07-12T19:37:48.367Z",
                "organization_id": "orgsUvmxKLxSZPtHH8",
            },
            201,
        )


class FakeOrganizations:

    def count(self, *args, **kwargs):
        return self.list()[0]["count"]

    def get(self, *args, **kwargs) -> QuerySet:
        return QuerySet(self, *args, **kwargs)

    def _get(self, *args, **kwargs):
        return self.list()[0]["results"]

    def list(self):
        return (
            {
                "count": 3,
                "next": None,
                "previous": None,
                "results": [
                    {
                        "id": "org_gxUvmxKLxSZPtHH8",
                        "name": "splight",
                        "display_name": "Splight",
                        "metadata": {
                            "billing_id": "123",
                            "blockchain_id": (
                                "0xad087302625aD69609248554B507dBE22B129F96"
                            ),
                        },
                    },
                    {
                        "id": "org_qgoGDtof5MnwzzMe",
                        "name": "organization_test",
                        "display_name": "Organization_test",
                        "metadata": {
                            "billing_id": None,
                            "blockchain_id": None,
                        },
                    },
                    {
                        "id": "org_HhMefMw2ioFGDJqw",
                        "name": "integration_splight",
                        "display_name": "Integration_splight",
                        "metadata": {
                            "billing_id": None,
                            "blockchain_id": None,
                        },
                    },
                ],
            },
            200,
        )

    def retrieve(self, resource_id: str):
        return (
            {
                "id": "org_HhMefMw2ioFGDJqw",
                "name": "integration_splight",
                "display_name": "Integration_splight",
                "metadata": {
                    "billing_id": None,
                    "blockchain_id": None,
                },
            },
            200,
        )

    def update(self, resource_id: str, data: Dict[str, Any]):
        return (
            {
                "id": "org_HhMefMw2ioFGDJqw",
                "name": "integration_splight",
                "display_name": "Integration_splight",
                "metadata": {
                    "billing_id": None,
                    "blockchain_id": None,
                },
            },
            200,
        )

    def invite_manager(self, organization_id: str, data: Dict[str, Any]):
        return ({}, 200)

    def create(self, data: Dict[str, Any]):
        logger.info("[FAKED] Created new org")
        data.update({"id": "org_HhMefMw2ioFGDJqw"})
        return data, 201

    def destroy(self, resource_id: str):
        logger.info("[FAKED] Deleted org {id}")
        pass


class FakeAuthClient(AbstractAuthClient):
    def __init__(self, *args, **kwargs):
        super().__init__()

        self._credentials = FakeCredentials()
        self._deployment = FakeDeployment()
        self._profile = FakeProfile()
        self._role = FakeRoles()
        self._user = FakeUsers()
        self._invitation = FakeInvitations()
        self._organization = FakeOrganization()
        self._organizations = FakeOrganizations()

    @property
    def credentials(self):
        return self._credentials

    @property
    def deployment(self):
        return self._deployment

    @property
    def profile(self):
        return self._profile

    @property
    def role(self):
        return self._role

    @property
    def user(self):
        return self._user

    @property
    def invitation(self):
        return self._invitation

    @property
    def organization(self):
        return self._organization

    @property
    def organizations(self):
        return self._organizations

    def authenticate(self):
        user_complete_info: Dict = {
            "organization_id": "org_gxUvmxKLxSZPtHH8",
            "original_organization_id": "org_gxUvmxKLxSZPtHH3",
            "user_id": "auth0|628f9e023211690069ccc193",
            "username": "fake.user",
            "name": "fake.user@splight-ae.com",
            "email": "fake.user@splight-ae.com",
            "picture": "https://s.gravatar.com/avatar/a7b54dab1cd2b75ec1b1a8f6ed31efdc?s=480&r=pg&d=https%3A%2F%2Fcdn.auth0.com%2Favatars%2Fma.png",
            "permissions": ["admin", "splightadmin"],
            "user_metadata": {
                "language": "en",
                "theme": "light",
                "picture_color": "light-green",
            },
            "app_metadata": {"is_manager": True},
            "role": "splightadmin",
        }
        user: User = User.parse_obj(user_complete_info)
        return (user, None)

    def authenticate_header(self):
        return "Bearer"



from pydantic import Field
from typing import Optional
from datetime import datetime, timezone
from splight_models.base import SplightBaseModel

class DatalakeModel(SplightBaseModel):
    timestamp: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))

    class Config:
        validate_assignment = True


class ComponentDatalakeModel(DatalakeModel):
    instance_id: Optional[str] = None
    instance_type: Optional[str] = None

    def dict(self, *args, **kwargs):
        d = super().dict(*args, **kwargs)
        return {
            k: v['id'] if isinstance(v, dict) else v
            for k, v in d.items()
        }



import os
import shutil
import pathlib
import zlib
from base64 import urlsafe_b64encode as b64e, urlsafe_b64decode as b64d
from pydantic import BaseModel
from typing import Optional, Type, List
from splight_lib import logging
from splight_models.storage import StorageFile
from splight_abstract import AbstractStorageClient
from splight_lib.settings import SPLIGHT_HOME


STORAGE_HOME = os.path.join(SPLIGHT_HOME, "storage")
logger = logging.getLogger()


class FakeStorageClient(AbstractStorageClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.base_path = os.path.join(STORAGE_HOME, self.namespace)
        os.makedirs(STORAGE_HOME, exist_ok=True)
        os.makedirs(self.base_path, exist_ok=True)

    def _encode_name(self, name):
        name = name.replace(f"{self.namespace}/", '')
        return b"fake" + b64e(zlib.compress(name.encode('utf-8'), 9))

    def _decode_name(self, name):
        name = name.replace("fake", "")
        decoded_name = zlib.decompress(b64d(name)).decode('utf-8')
        return f"{self.namespace}/{decoded_name}"

    def _namespaced_key(self, name):
        if not name.startswith(self.namespace):
            name = f"{self.namespace}/{name}"
        return name

    def copy(self, old_name, new_name):
        new_name_path = os.path.split(new_name)
        if new_name_path[0] != "":
            os.makedirs(os.path.join(*new_name_path[:-1]), exist_ok=True)
        try:
            shutil.copy(old_name, new_name)
        except shutil.SameFileError:
            logger.exception("Source and destination represents the same file.")
        except PermissionError:
            logger.exception("Permission denied.")
        except:
            logger.exception("Error occurred while copying file.")

    def save(self, instance: BaseModel, prefix: Optional[str] = None) -> BaseModel:
        logger.debug(f"[FAKED] created file {instance}")
        id = os.path.split(instance.file)[-1]
        if prefix:
            if prefix.startswith('/'):
                prefix = pathlib.Path(prefix).relative_to('/')
            id = os.path.join(prefix, id)
        destination = os.path.join(STORAGE_HOME, self.namespace, id)
        self.copy(instance.file, destination)
        id = self._encode_name(id)
        instance.id = id
        return instance

    def _get(self,
             resource_type: Type,
             first=False,
             limit_: int = -1,
             skip_: int = 0,
             prefix: Optional[str] = None,
             **kwargs) -> List[BaseModel]:
        logger.debug(f"[FAKED] Getting files {kwargs}")
        base_path = self.base_path + "/" + prefix if prefix else self.base_path
        queryset = []
        if os.path.isfile(base_path):
            relative_path = str(pathlib.Path(base_path).relative_to(self.base_path))
            queryset = [StorageFile(id=self._encode_name(relative_path), file=relative_path)]
        else:
            for folder, _, files in os.walk(base_path):
                for file in files:
                    queryset.append(os.path.join(folder, file))
            queryset = [str(pathlib.Path(f).relative_to(self.base_path)) for f in queryset]
            queryset = [StorageFile(id=self._encode_name(f), file=f) for f in queryset]
        kwargs = self._validated_kwargs(resource_type, **kwargs)
        queryset = self._filter(queryset, **kwargs)
        if limit_ != -1:
            queryset = queryset[skip_:skip_ + limit_]

        if first:
            return queryset[0] if queryset else None
        return queryset

    def delete(self, resource_type: Type, id: str) -> List[BaseModel]:
        logger.debug(f"[FAKED] Deleted file {id}")
        id = self._namespaced_key(self._decode_name(id))
        os.remove(os.path.join(STORAGE_HOME, id))

    def download(self, resource_type: Type, id: str, target: str) -> str:
        logger.debug(f"[FAKED] Downloading file {id} to {target}")
        id = self._namespaced_key(self._decode_name(id))
        source = os.path.join(STORAGE_HOME, id)
        self.copy(source, target)
        return target

    def upload(self, id: str):
        raise NotImplementedError

    def get_temporary_link(self, instance: BaseModel) -> str:
        raise NotImplementedError



from splight_models.namespace import Namespace
from splight_lib import logging
from splight_lib.settings import setup


DatabaseClient = setup.DATABASE_CLIENT
DeploymentClient = setup.DEPLOYMENT_CLIENT
logger = logging.getLogger()


class OrganizationHandler:

    @staticmethod
    def create(id: str):
        id = str(id)
        # TODO add Auth0 step
        namespace = Namespace(id=id)
        DeploymentClient().save(instance=namespace)
        return DatabaseClient().save(namespace)
    
    @staticmethod
    def delete(id):
        DeploymentClient().delete(Namespace, id=id)
        return DatabaseClient().delete(Namespace, id=id) # TODO logic delete



from datetime import datetime
from unittest import TestCase

from splight_models import *
from splight_models.base import SplightBaseModel


class AbstractBaseHash(object):
    class_model = SplightBaseModel
    args = {}
    rand_str_arg = ""

    def test_eq_hash(self):
        obj1 = self.class_model(**self.args)
        obj2 = self.class_model(**self.args)
        self.assertEqual(obj1.__hash__(), obj2.__hash__())

    def test_neq_hash(self):
        obj1_args = self.args.copy()
        obj2_args = self.args.copy()
        obj1_args[self.rand_str_arg] = "randstr1"
        obj2_args[self.rand_str_arg] = "randstr2"
        obj1 = self.class_model(**obj1_args)
        obj2 = self.class_model(**obj2_args)
        self.assertNotEqual(obj1.__hash__(), obj2.__hash__())


class TestAssetHash(AbstractBaseHash, TestCase):
    class_model = Asset
    args = {"name": ""}
    rand_str_arg = "name"


class TestAttributeHash(AbstractBaseHash, TestCase):
    class_model = Attribute
    args = {"name": ""}
    rand_str_arg = "name"


class TestDeploymentHash(AbstractBaseHash, TestCase):
    class_model = Deployment
    args = {"version": "ver", "type": "Component"}
    rand_str_arg = "version"


class TestGeopointHash(AbstractBaseHash, TestCase):
    class_model = Geopoint
    args = {"id": "id", "latitude": 0.1, "longitude": 0.2}
    rand_str_arg = "id"


class TestMappingHash(AbstractBaseHash, TestCase):
    class_model = Mapping
    args = {
        "name": "MyMapping",
        "asset_id": "1",
        "attribute_id": "1",
        "output_format": "Number",
    }
    rand_str_arg = "asset_id"


class TestMessageHash(AbstractBaseHash, TestCase):
    class_model = Message
    args = {"action": "1", "variables": []}
    rand_str_arg = "action"


class TestNamespaceHash(AbstractBaseHash, TestCase):
    class_model = Namespace
    args = {"id": "1"}
    rand_str_arg = "id"


class TestNetTargetHash(AbstractBaseHash, TestCase):
    class_model = NetTarget
    args = {"ip": "1", "port": "1"}
    rand_str_arg = "ip"


class TestNotificationHash(AbstractBaseHash, TestCase):
    class_model = Notification
    args = {"title": "1", "message": "1", "timestamp": datetime.now()}
    rand_str_arg = "title"

class TestComponentHash(AbstractBaseHash, TestCase):
    class_model = Component
    args = {"name": "1", "version": "1"}
    rand_str_arg = "name"

class TestStorageFileHash(AbstractBaseHash, TestCase):
    class_model = StorageFile
    args = {"file": "1"}
    rand_str_arg = "file"


class TestTagHash(AbstractBaseHash, TestCase):
    class_model = Tag
    args = {"name": "1"}
    rand_str_arg = "name"


class TestVariableHash(AbstractBaseHash, TestCase):
    class_model = Variable
    args = {"path": "1", "args": {}, "timestamp": datetime.now()}
    rand_str_arg = "path"



from splight_lib.shortcut.notification import notify
from splight_lib.shortcut.file import save_file
from splight_lib.shortcut.organization import OrganizationHandler

__all__ = [
    notify,
    save_file,
    OrganizationHandler,
]



import time
import atexit
import sys
import uuid
from typing import List, Callable, Union, Tuple, Any
from threading import (
    Thread as DefaultThread,
    Event,
    Lock,
)
from subprocess import Popen as DefaultPopen
from functools import wraps
from splight_abstract.client.abstract import AbstractClient
from splight_lib import logging


logger = logging.getLogger()


class Empty(object):
    pass


class Task:
    """A periodic task for the scheduler."""

    def __init__(self, handler: Callable[[Any], None], args: Tuple, period: int, hash: str = None) -> None:
        if hash is None:
            hash = str(uuid.uuid4())
        self.hash = hash
        self.args = args
        self.period = period
        self.handler = handler
        self._name = self.handler

    def __repr__(self):
        return "<%s(%s)>" % (self.__class__.__name__, self._name)


class TaskSet:
    """Set of Tasks with the same hash but different or equal periods."""

    def __init__(self, task: Task) -> None:
        self.hash: str = task.hash
        self.args = task.args
        self.handler = task.handler
        self._periods: List[int] = [task.period]
        self.next_event = time.time()

    def update_event(self) -> None:
        """Set next_event of the TaskSet."""
        if len(self._periods) > 0:
            self.next_event += min(self._periods)
        else:
            # 2**16 is enough to wait and don't overflow.
            self.next_event = time.time() + 2 ** 16

    def in_time(self, now: float) -> bool:
        """
        Check if the TaskSet is in time to execute.

        Args:
            now (int): The time now.
            return (bool): Return True if the TaskSet is in time.
        """
        return self.next_event <= now

    def add_period(self, period: float) -> bool:
        """
        Add new task period to the set. Update next event if is needed.

        Args:
            period (float): New scan period to add.
            return (bool): Return True if the task needs to
                be recalculated.
        """
        self._periods.append(period)

        # If is the first period, recalculate the next_event
        if len(self._periods) <= 1:
            self.next_event = time.time()
            return True

        return period == min(self._periods)

    def remove_period(self, period: float) -> bool:
        """
        Remove a task period to the list. Update execute if is needed.

        Args:
            period (float): Scan period to remove.
            return (bool): Return True if the next_event needs to
                be recalculated.
        """
        if period not in self._periods:
            return False

        # If is the min, you has to run the scheduler
        ret: bool = period == min(self._periods)

        # Precondition: period in self._periods
        self._periods.remove(period)

        return ret


class TaskMap:
    """
    Class used to manage tasks on the scheduler.
    """

    def __init__(self) -> None:
        self._map: dict[str, Task] = {}

    def add_task(self, task: Task) -> bool:
        """
        Add new task to the map.

        Args:
            task (Task): Task to add.
            return (bool): Return True if a scheduler run is needed.
        """
        if task.hash not in self._map:
            self._map[task.hash] = TaskSet(task)
            return True

        return self._map[task.hash].add_period(task.period)

    def remove_task(self, task: Task) -> bool:
        """
        Add new task to the map.

        Args:
            task (Task): Task to removed.
            return (bool): Return True if a scheduler run is needed.

        Note: The period is needed to prevent deleting other scan.
        """
        if task.hash not in self._map:
            return False

        # This don't delete the key if is empty, it does not bother
        return self._map[task.hash].remove_period(task.period)

    def get_list(self) -> List[TaskSet]:
        """Get the list of scans."""
        return list(self._map.values())


class Scheduler:
    def __init__(self, *args, **kwargs) -> None:
        self._tasks = TaskMap()
        self._event = Event()
        self._mutex = Lock()
        self._to_add: List[Task] = []
        self._to_remove: List[Task] = []

    def start(self) -> None:
        """Scheduler infinite loop."""
        self._stop = False
        while not self._stop:
            # Update the task list
            self._update_task_list()
            # Run the scheduler task
            near_event = self._schedule()
            # Get the next event
            next_event_time = near_event - time.time()

            # Wait until next event or someone trigger the event
            self._event.wait(timeout=next_event_time)
            self._event.clear()

    def stop(self) -> None:
        self._stop = True

    def _schedule(self) -> float:
        """Scheduler main task."""
        now = time.time()
        near_event: float = now + 65536  # 2**16

        for task_set in self._tasks.get_list():
            if task_set.in_time(now):
                task_set.handler(*task_set.args)
                task_set.update_event()

            near_event = min(near_event, task_set.next_event)

        return near_event

    def _notify_scheduler(self) -> None:
        """
        Unlock the scheduler if is locked.
        This will run the scheduler main task.
        """
        self._event.set()

    def schedule(self, task: Task) -> None:
        """
        Add new task to the scheduler.

        Args:
            task (Task): The new job.
        """

        with self._mutex:
            self._to_add.append(task)

        self._notify_scheduler()

    def unschedule(self, task: Task) -> None:
        """
        Delete a task from the scheduler.

        Args:
            task (Task): Job to delete.
        """
        with self._mutex:
            self._to_remove.append(task)

        self._notify_scheduler()

    def _update_task_list(self) -> None:
        """
        Update the task list.
        """
        with self._mutex:
            for task in self._to_add:
                self._tasks.add_task(task)
            for task in self._to_remove:
                self._tasks.remove_task(task)

            self._to_add = []
            self._to_remove = []


class Thread(DefaultThread):
    def __init__(self, target: Callable, args: Tuple = (), **kwargs) -> None:
        target = self.store_result(target)
        self.result = Empty()
        super().__init__(target=target, args=args, name=target, **kwargs)

    def store_result(self, func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs):
            self.result = func(*args, **kwargs)
            return self.result
        return wrapper

    def exit_ok(self) -> bool:
        return not isinstance(self.result, Empty)

    def kill(self) -> None:
        pass

    def terminate(self) -> None:
        pass


class Popen(DefaultPopen):

    def __init__(self, args: List[str]) -> None:
        self.args: List[str] = args
        super(Popen, self).__init__(args)

    def is_alive(self) -> bool:
        return self.poll() is None

    def exit_ok(self) -> bool:
        return self.poll() == 0


class ExecutionClient(AbstractClient):
    def __init__(self, *args, **kwargs):
        self.processes: List[Thread] = []
        self.threads: List[Popen] = []

        self._register_exit_functions()
        super(ExecutionClient, self).__init__(*args, **kwargs)

    def _register_exit_functions(self) -> None:
        excepthook = sys.excepthook

        def wrap_excepthook(type, value, traceback):
            self.terminate_all()
            excepthook(type, value, traceback)

        atexit.register(self.terminate_all)
        sys.excepthook = wrap_excepthook

    def __del__(self) -> None:
        self.terminate_all()

    def terminate_all(self) -> None:
        for p in self.processes:
            p.terminate()

    def start(self, job=Union[Popen, Thread, Task]):
        if isinstance(job, Popen):
            return self._start_process(job)
        if isinstance(job, Thread):
            return self._start_thread(job)
        if isinstance(job, Task):
            return self._start_task(job)

    def stop(self, job=Union[Popen, Thread, Task]):
        if isinstance(job, Popen):
            raise NotImplementedError
        if isinstance(job, Thread):
            raise NotImplementedError
        if isinstance(job, Task):
            return self._stop_task(job)

    def _start_process(self, job: Popen) -> None:
        logger.debug(f"Starting process {job}")
        self.processes.append(job)
        return

    def _start_thread(self, job: Thread) -> None:
        logger.debug(f"Starting Thread {job}")
        self.threads.append(job)
        job.start()
        return

    def _start_task(self, job: Task) -> None:
        logger.debug(f"Starting Task {job}")
        if not getattr(self, '_scheduler', None):
            # Instantiate and start Scheduler thread
            self._scheduler = Scheduler()
            self._start_thread(Thread(target=self._scheduler.start))

        return self._scheduler.schedule(job)

    def _stop_task(self, job: Task) -> None:
        logger.debug(f"Stopping Task {job}")
        return self._scheduler.unschedule(job)

    def healthcheck(self):
        return all([p.is_alive() or p.exit_ok() for p in self.processes + self.threads])



from .base import SplightBaseModel
from typing import List


class NetTarget(SplightBaseModel):
    ip: str
    port: int


class NetRule(SplightBaseModel):
    origin: NetTarget
    destination: NetTarget


class NetValue(SplightBaseModel):
    value: List[NetRule]




from abc import ABC, abstractmethod


class AbstractAuthClient(ABC):
    @property
    @abstractmethod
    def credentials(self):
        pass

    @property
    @abstractmethod
    def profile(self):
        pass

    @property
    @abstractmethod
    def role(self):
        pass

    @property
    @abstractmethod
    def user(self):
        pass

    @property
    @abstractmethod
    def organization(self):
        pass

    @property
    @abstractmethod
    def organizations(self):
        pass

    @abstractmethod
    def authenticate(self):
        pass

    @abstractmethod
    def authenticate_header(self):
        pass



from unittest import TestCase
from splight_lib.execution import ExecutionClient, Task, Thread, Popen
from pathlib import Path
import time
import os


class TestExecutionClient(TestCase):
    def setUp(self) -> None:
        self.base_dir = Path(__file__).resolve().parent.parent
        self.sleep_time = 0.25
        self.executions = []
        return super().setUp()

    @staticmethod
    def function_ok() -> None:
        pass

    @staticmethod
    def function_nok() -> None:
        raise Exception('Test exception')

    def function_to_schedule(self, arg1, arg2) -> None:
        self.executions.append(('function_to_schedule', arg1, arg2))

    def test_thread_healthcheck_ok(self) -> None:
        client = ExecutionClient()
        client.start(Thread(self.function_ok))
        time.sleep(self.sleep_time)
        self.assertTrue(client.healthcheck())

    def test_thread_healthcheck_fail(self) -> None:
        client = ExecutionClient()
        client.start(Thread(self.function_nok))
        time.sleep(self.sleep_time)
        self.assertFalse(client.healthcheck())

    def test_process_healthcheck_ok(self) -> None:
        client = ExecutionClient()
        file_path = os.path.join(self.base_dir, 'tests/FakeProc.py')
        client.start(Popen(["python", file_path, 'exit_ok']))
        time.sleep(self.sleep_time)
        self.assertTrue(client.healthcheck())

    def test_process_healthcheck_fail(self) -> None:
        client = ExecutionClient()
        file_path = os.path.join(self.base_dir, 'tests/FakeProc.py')
        client.start(Popen(["python", file_path, 'exit_fail']))
        time.sleep(self.sleep_time)
        self.assertFalse(client.healthcheck())

    def test_terminate_all(self) -> None:
        client = ExecutionClient()
        file_path = os.path.join(self.base_dir, 'tests/FakeProc.py')
        client.start(Popen(["python", file_path, 'run_forever']))
        time.sleep(self.sleep_time)
        self.assertTrue(client.healthcheck())
        client.terminate_all()
        time.sleep(self.sleep_time)
        self.assertFalse(client.processes[0].is_alive())

    def test_scheduled_task(self) -> None:
        client = ExecutionClient()
        task = Task(handler=self.function_to_schedule, args=("arg1", 2), period=1, hash="WTF")
        self.assertEqual(len(self.executions), 0)
        # Start task
        client.start(task)

        # Check 1 more execution in a period after start
        prev_count = len(self.executions)
        time.sleep(0.99)
        self.assertEqual(len(self.executions), prev_count + 1)
        self.assertEqual(self.executions[-1], ("function_to_schedule", "arg1", 2))

        # Stop task
        client.stop(task)
        client._scheduler.stop()
        time.sleep(0.5)

        # Check no more executions after stop
        prev_count = len(self.executions)
        time.sleep(3)
        self.assertEqual(len(self.executions), prev_count)



import re
from typing import Dict, Optional
from splight_abstract import AbstractCacheClient


class FakeCacheClient(AbstractCacheClient):

    data: Dict[str, str] = dict()

    def __init__(self, *args, **kwargs):
        super(FakeCacheClient, self).__init__(*args, **kwargs)

    def get(self, key) -> Optional[bytes]:
        val = self.data.get(key, None)
        return val.encode("utf-8") if val else None

    def set(self, key, value, *args, **kwargs):
        self.data[key] = value

    def delete(self, pattern):
        pattern = pattern.replace('*', ".*")
        new_dict = {}
        for key in self.data:
            if not re.match(f'{pattern}', key):
                new_dict[key] = self.data[key]
        self.data = new_dict

    def clear(self):
        self.data.clear()



from abc import ABC, abstractmethod, abstractproperty
from typing import Dict, List, Type, Tuple

from pydantic import BaseModel

from splight_abstract.client import AbstractClient, QuerySet


class AbstractHubSubClient(AbstractClient):
    @abstractmethod
    def save(self, instance: BaseModel) -> BaseModel:
        pass

    def get(self, *args, **kwargs) -> QuerySet:
        return QuerySet(self, *args, **kwargs)

    @abstractmethod
    def _get(self, resource_type: Type, first=False, limit_: int = -1, skip_: int = 0, **kwargs) -> List[BaseModel]:
        pass

    @abstractmethod
    def delete(self, resource_type: Type, id: str) -> None:
        pass

    @abstractmethod
    def update(self, resource_type: Type, id: str, data: Dict) -> BaseModel:
        pass


class AbstractHubClient(ABC):
    @abstractmethod
    def upload(self, data: Dict, files: Dict) -> Tuple:
        pass

    @abstractmethod
    def download(self, data: Dict) -> Tuple:
        pass

    @abstractmethod
    def random_picture(self):
        pass

    @abstractproperty
    def all(self) -> AbstractHubSubClient:
        pass

    @abstractproperty
    def mine(self) -> AbstractHubSubClient:
        pass

    @abstractproperty
    def public(self) -> AbstractHubSubClient:
        pass

    @abstractproperty
    def private(self) -> AbstractHubSubClient:
        pass

    @abstractproperty
    def setup(self) -> AbstractHubSubClient:
        pass

    @abstractproperty
    def system(self) -> AbstractHubSubClient:
        pass



from pydantic import BaseModel


class SplightBaseModel(BaseModel):
    def __hash__(self):  # make hashable BaseModel subclass
        return hash(self.__class__.__name__ + str(tuple(self.dict().values())))

    @staticmethod
    def get_event_name(type: str, action: str) -> str:
        return f"{type.lower()}_{action}"



from splight_abstract import AbstractBlockchainClient, AbstractRemoteClient
from remote_splight_lib.auth import SplightAuthToken
from remote_splight_lib.settings import settings
from splight_models import Transaction, BlockchainContract
from furl import furl
from requests import Session
from typing import Dict
from hexbytes import HexBytes
import json


class BlockchainClient(AbstractBlockchainClient, AbstractRemoteClient):
    _PREFIX = "blockchain"

    def __init__(self, namespace: str = "default") -> None:
        super().__init__()
        super(BlockchainClient, self).__init__(namespace=namespace)
        self._base_url = furl(settings.SPLIGHT_PLATFORM_API_HOST)
        token = SplightAuthToken(
            access_key=settings.SPLIGHT_ACCESS_ID,
            secret_key=settings.SPLIGHT_SECRET_KEY,
        )
        self._session = Session()
        self._session.headers.update(token.header)

    def transact(self, method: str, private_key: str, contract_address: str, **kwargs) -> Dict:
        # POST /blockchain/contract/{contract.address}/transact/{method}
        url = self._base_url / f"{self._PREFIX}/contract/{contract_address}/transact/{method}/"
        data = {"private_key": private_key, **kwargs}
        response = self._session.post(url, json=data)
        response.raise_for_status()
        return response

    def call(self, method: str, *args) -> Dict:
        return super().call(method, *args)

    def get_balance(self) -> int:
        return super().get_balance()

    def get_transaction(self, tx_hast: HexBytes) -> Dict:
        return super().get_transaction(tx_hast)



import json
from typing import Callable, Dict

from splight_lib.logging import logging
from splight_abstract.communication import AbstractCommunicationClient, CommunicationContext
from splight_models.communication import CommunicationEvent


logger = logging.getLogger()


class FakeCommunicationClient(AbstractCommunicationClient):
    @property
    def context(self):
        return CommunicationContext(
            auth_headers=None,
            auth_endpoint=None,
            key='key',
            channel='default',
            channel_data=None
        )

    @property
    def status(self):
        return 'ready'

    @property
    def channel(self):
        return 'channel'

    def unbind(self, event_name: str, event_handler: Callable):
        logger.debug(f"[FAKED] unbind {event_name} {event_handler}")

    def bind(self, event_name: str, event_handler: Callable):
        logger.debug(f"[FAKED] bind {event_name} {event_handler}")

    def trigger(self, event: CommunicationEvent):
        logger.debug(f"[FAKED] trigger {event.event_name} {event.data} {event.socket_id} {event.instance_id}")

    def authenticate(self, channel_name: str, socket_id: str, custom_data: Dict = None) -> Dict:
        logger.debug(f"[FAKED] authenticate {channel_name} {socket_id} {custom_data}")
        return {
            "auth": "asd",
            "socket_id": socket_id,
            "channel_data": json.dumps(custom_data),
            "channel_name": channel_name
        }



from splight_abstract import AbstractNotificationClient
from splight_models import Notification
from splight_lib import logging


logger = logging.getLogger()


class FakeNotificationClient(AbstractNotificationClient):
    @property
    def channel(self):
        return f"private-{self.namespace}"

    def send(self, instance: Notification, topic: str = 'default') -> None:
        logger.info(f"[FAKED] Notification sent channel:{self.channel} topic:{topic} data:{instance.dict()} ")


from .base import SplightBaseModel
from typing import Optional, Dict
from pydantic import validator
import json
import os


class File(SplightBaseModel):
    id: Optional[str] = None
    file: str
    description: Optional[str] = None
    metadata: Dict = {}
    content_type: Optional[str] = None
    url: Optional[str] = None

    @validator("file", pre=True)
    def validate_file(cls, v):
        v = v.replace('/', os.sep)
        v = v.replace('\\', os.sep)
        return v

    @property
    def name(self):
        return self.file.split(os.sep)[-1]

    def json(self, *args, **kwargs):
        prev_metadata = self.metadata
        self.metadata = json.dumps(self.metadata)
        res = super(File, self).json(*args, **kwargs)
        self.metadata = prev_metadata
        return res



from .base import SplightBaseModel
from typing import List, Optional


class Asset(SplightBaseModel):
    id: Optional[str]
    name: str
    description: Optional[str] = None
    external_id: Optional[str]
    tags: List[str] = []



import json
from functools import wraps
from typing import Callable, List
from abc import abstractmethod
from splight_abstract.client import AbstractClient
from splight_lib import logging


logger = logging.getLogger()


class AbstractCacheClient(AbstractClient):

    @abstractmethod
    def get(self, key):
        pass

    @abstractmethod
    def set(self, key, value):
        pass

    @abstractmethod
    def delete(self, key):
        pass

    @abstractmethod
    def clear(self):
        pass

# Decorators
def get_cache(client: AbstractCacheClient, cache_key: str, cache_key_args: str or List[str] = None) -> Callable:
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs) -> Callable:
            final_cache_key = cache_key
            cached_data = None
            full_args = dict(zip(func.__code__.co_varnames, args))
            full_args.update(kwargs)
            if cache_key_args is not None:
                if isinstance(cache_key_args, str):
                    final_cache_key = f'{final_cache_key}:{cache_key_args}-{str(full_args.get(cache_key_args))}'
                else:
                    for key_arg in cache_key_args:
                        final_cache_key = f'{final_cache_key}:{key_arg}-{str(full_args.get(key_arg))}'

            try:
                cached_data = client.get(final_cache_key)
            except Exception:
                logger.error(f'Error getting cache for {final_cache_key}')

            if isinstance(cached_data, bytes):
                logger.info(f'Cache hit for {final_cache_key}')

                return json.loads(cached_data.decode('utf-8'))
            else:
                cached_data = func(*args, **kwargs)

                try:
                    client.set(final_cache_key, json.dumps(cached_data), ex=3600)
                    logger.info(f'Cache update for {final_cache_key}')
                except Exception:
                    logger.error(f'Error getting cache for {final_cache_key}')

                return cached_data
        return wrapper
    return decorator


def flush_cache(client: AbstractCacheClient, cache_key: str, cache_key_args: str or List[str] = None, prefix_match=False) -> Callable:
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs) -> Callable:
            final_cache_key = cache_key
            full_args = dict(zip(func.__code__.co_varnames, args))
            full_args.update(kwargs)
            if cache_key_args is not None:
                if isinstance(cache_key_args, str):
                    final_cache_key = f'{final_cache_key}:{cache_key_args}-{str(full_args.get(cache_key_args))}'
                else:
                    for key_arg in cache_key_args:
                        final_cache_key += f':{key_arg}-{str(full_args.get(key_arg))}'

            try:
                if prefix_match:
                    final_cache_key += '*'
                client.delete(final_cache_key)
                logger.info(f'Cache flush for {final_cache_key}')
            except Exception:
                logger.error(f'Error flushing cache for {final_cache_key}')

            return func(*args, **kwargs)
        return wrapper
    return decorator



from enum import Enum
from typing import Dict, Optional
from splight_models import SplightBaseModel
from splight_models.user import User


class CommunicationChannelData(SplightBaseModel):
    user_id: str
    user_info: Dict

    @classmethod
    def parse_from_user(cls, user: User):
        return cls.parse_obj(
            {
                "user_id": user.user_id,
                "user_info": user.dict()
            }
        )

class CommunicationContext(SplightBaseModel):
    auth_headers: Optional[Dict] = None
    auth_endpoint: Optional[str] = None
    key: str
    channel: str
    private_room_channel: str
    presence_room_channel: str
    channel_data: Optional[CommunicationChannelData] = None


class CommunicationClientStatus(str, Enum):
    STOPPED = 'stopped'
    STARTING = 'starting'
    READY = 'ready'
    FAILED = 'failed'
    ERROR = 'error'



from splight_models import *
from splight_lib import logging
from splight_lib.settings import setup


NotificationClient = setup.NOTIFICATION_CLIENT
DatalakeClient = setup.DATALAKE_CLIENT
DatabaseClient = setup.DATABASE_CLIENT
logger = logging.getLogger()


def notify(notification: Notification,
           database_client: DatabaseClient = None,
           notification_client: NotificationClient = None,
           datalake_client: DatalakeClient = None) -> None:
    logger.debug(f"Notification sent: {notification}")
    if database_client:
        notification = database_client.save(notification)
    if notification_client:
        notification_client.send(notification)
    if datalake_client:
        variable = Variable(
            timestamp=notification.timestamp,
            args=notification.dict(),
        )
        datalake_client.save(instances=[variable], collection='notification')
    return



from splight_models import (
    Asset,
    Attribute,
    File,
    Mapping,
    Edge,
    Graph,
    Component,
    Notification,
    Query,
    Node,
    Tag,
    BlockchainContract,
    BlockchainContractSubscription,
    Component,
    ComponentObject,
    ComponentCommand,
)


CLASSMAP = {
    Component: {
        "path": "component"
    },
    Asset: {
        "path": "asset"
    },
    Attribute: {
        "path": "attribute"
    },
    Mapping: {
        "path": "mapping"
    },
    Edge: {
        "path": "graph/edge"
    },
    File: {
        "path": "file"
    },
    Graph: {
        "path": "graph/graph"
    },
    # Namespace: {
    #     "path": "asset"
    # },
    Notification: {
        "path": "notification"
    },
    Node: {
        "path": "graph/node"
    },
    Tag: {
        "path": "tag"
    },
    BlockchainContract: {
        "path": "blockchain/contract"
    },
    BlockchainContractSubscription: {
        "path": "blockchain/subscription"
    },
    Component: {
        "path": "components"
    },
    ComponentObject: {
        "path": "component-objects"
    },
    ComponentCommand: {
        "path": "component-commands"
    },
    Query: {
        "path": "query"
    },
}



from splight_models.asset import *
from splight_models.attribute import *
from splight_models.blockchain import *
from splight_models.communication import *
from splight_models.channel import *
from splight_models.credential import *
from splight_models.datalake import *
from splight_models.deployment import *
from splight_models.file import *
from splight_models.geopoint import *
from splight_models.graph import *
from splight_models.hub import *
from splight_models.mapping import *
from splight_models.message import *
from splight_models.namespace import *
from splight_models.network import *
from splight_models.notification import *
from splight_models.tag import *
from splight_models.query import *
from splight_models.component import *
from splight_models.storage import *
from splight_models.user_activity import *
from splight_models.tag import *
from splight_models.user import *
from splight_models.variable import *



from unittest import TestCase
from parameterized import parameterized
from splight_lib.shortcut import notify
from fake_splight_lib.database import FakeDatabaseClient as DatabaseClient
from fake_splight_lib.datalake import FakeDatalakeClient as DatalakeClient
from fake_splight_lib.notification import FakeNotificationClient as NotificationClient
from splight_models.notification import SourceType
from splight_models import Notification, Variable


class TestNotification(TestCase):
    def setUp(self):
        self.namespace = "default"
        self.database = DatabaseClient(self.namespace)
        self.datalake = DatalakeClient(self.namespace)
        self.notifier = NotificationClient(self.namespace)
        return super().setUp()

    @parameterized.expand([
        (Notification(
            message="Message sample",
            title="Title sample"
        ),),
        (Notification(
            message="Message sample",
            title="Title sample",
            asset_id=1,
            attribute_id=1,
        ),),
        (Notification(
            message="Message sample",
            title="Title sample",
            asset_id=1,
            attribute_id=1,
            source_id='3',
        ),),
    ])
    def test_notify(self, notification):
        notify(
            notification=notification,
            database_client=self.database,
            datalake_client=self.datalake,
            notification_client=self.notifier)
        self.assertEqual(len(self.database.get(Notification, id=notification.id)), 1)
        self.assertEqual(self.database.get(Notification, id=notification.id)[0], notification)
        self.assertEqual(len(self.datalake.get(Variable, args__id=notification.id, collection='notification')), 1)
        self.assertEqual(self.datalake.get(Variable, args__id=notification.id, collection='notification')[0].args, notification.dict())
        self.assertEqual(self.datalake.get(Variable, args__id=notification.id, collection='notification')[0].timestamp, notification.dict().get('timestamp'))



from pydantic import Field
from datetime import datetime, timezone
from typing import Optional
from enum import Enum
from splight_models.datalake import DatalakeModel
from splight_models.severity import SeverityType


SYSTEM = 'system'
INFO = 'info'
LOW = 'low'
MEDIUM = 'medium'
HIGH = 'high'
CRITICAL = 'critical'

GREATER_THAN = 'gt'
GREATER_THAN_OR_EQUAL = 'ge'
LOWER_THAN = 'lt'
LOWER_THAN_OR_EQUAL = 'le'
EQUAL = 'eq'

SEVERITIES = (
    (SYSTEM, 'system'),
    (INFO, 'info'),
    (LOW, 'low'),
    (MEDIUM, 'medium'),
    (HIGH, 'high'),
    (CRITICAL, 'critical'),
)

SOURCE_TYPE = (
    ('Component', 'Component'),
)


class SourceType(str, Enum):
    # TODO: Remove this once component types are unified
    component = 'Component'
    algorithm = 'Algorithm'
    network = 'Network'
    connector = 'Connector'


class Notification(DatalakeModel):
    id: Optional[str]
    title: str
    message: str
    seen: bool = False
    timestamp: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))
    severity: SeverityType = SeverityType.info
    source_id: Optional[str]
    source_type: Optional[SourceType] = SourceType.component



import pandas as pd
import operator
import os
import json
from collections import defaultdict
from collections.abc import MutableMapping
from datetime import datetime, timezone, timedelta
from pydantic import BaseModel
from typing import Dict, List, Type, Any, Union
from splight_models import VariableDataFrame, Query
from splight_abstract.datalake import AbstractDatalakeClient, validate_resource_type
from splight_lib import logging
from splight_lib.settings import SPLIGHT_HOME


DATALAKE_HOME = os.path.join(SPLIGHT_HOME, "datalake")
logger = logging.getLogger()


def flatten_dict(d, parent_key='', sep='__'):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


class FakeDatalakeClient(AbstractDatalakeClient):
    extra_operators = {
        "in": lambda x, y: x in y,
        "contains": lambda x, y: y in x,
        "gte": lambda x, y: x >= y,
        "lte": lambda x, y: x <= y,
    }

    def __init__(self, namespace: str = 'default', *args, **kwargs) -> None:
        super(FakeDatalakeClient, self).__init__(namespace, *args, **kwargs)
        self.collections = defaultdict(list)
        self._default_load()

    @staticmethod
    def _write_to_collection(collection: str, data: List[dict]) -> None:
        os.makedirs(DATALAKE_HOME, exist_ok=True)
        col_file = os.path.join(DATALAKE_HOME, collection)
        _prev_data = []
        if os.path.exists(col_file):
            with open(col_file, 'r+') as f:
                content = f.read()
                _prev_data = json.loads(content) if content else []
        data = _prev_data + data
        with open(col_file, 'w+') as f:
            f.write(json.dumps(data, indent=4, sort_keys=True, default=str))

    @staticmethod
    def _date_hook(json_dict):
        timeformat = "%Y-%m-%d %H:%M:%S.%f%z"
        for (key, value) in json_dict.items():
            try:
                json_dict[key] = datetime.strptime(value, timeformat)
            except:
                pass
        return json_dict

    @staticmethod
    def _read_from_collection(collection: str) -> List[dict]:
        os.makedirs(DATALAKE_HOME, exist_ok=True)
        col_file = os.path.join(DATALAKE_HOME, collection)
        content = ''
        if os.path.exists(col_file):
            with open(col_file, 'r+') as f:
                content = f.read()
        return json.loads(content, object_hook=FakeDatalakeClient._date_hook) if content else []

    @staticmethod
    def _resolve_filter(keys: List[str], value: Any, oper: str):
        def inner(x: Dict) -> bool:
            for key in keys:
                if not isinstance(x, dict) or not key in x:
                    return False
                x = x[key]

            if oper in FakeDatalakeClient.extra_operators:
                return FakeDatalakeClient.extra_operators[oper](x, value)
            return getattr(operator, oper)(x, value)

        return inner

    def _default_load(self):
        data = [
            {
                "_id": "jashdasd",
                "asset_id": "1",
                "attribute_id": "1",
                "args": {
                    "value": 1
                },
                "timestamp": datetime.now(timezone.utc)
            },
            {
                "_id": "jashdasd",
                "asset_id": "1",
                "attribute_id": "1",
                "args": {
                    "value": 4
                },
                "timestamp": datetime.now(timezone.utc)
            }
        ]
        self._write_to_collection('default', data)

    def _find(self, collection: str, filters: List = [], **kwargs) -> List[Dict]:
        values: List[Dict] = self._read_from_collection(collection)
        values = [v for v in values if all(f(v) for f in filters)]
        return values

    def list_collection_names(self):
        return [f for f in os.listdir(DATALAKE_HOME) if os.path.isfile(os.path.join(DATALAKE_HOME, f))]

    def get_unique_keys(self, collection: str, **kwargs):
        read = self._find(collection, filters=self._parse_filters(**kwargs))
        # flatten all dicts
        read = [flatten_dict(d) for d in read]
        return list(set(key for dic in read for key in dic.keys()))

    def get_values_for_key(self, collection: str, key: str, **kwargs):
        read = self._find(collection, filters=self._parse_filters(**kwargs))
        # flatten all dicts
        read = [flatten_dict(d) for d in read]
        ret = list(set(d[key] for d in read if key in d))
        return ret

    def _parse_filters(self, **kwargs) -> List:
        filters: List = []
        for key, value in kwargs.items():
            args: List[str] = key.split('__')
            oper = args[-1] if args[-1] in self.valid_filters else 'eq'
            args = args if oper == 'eq' else args[:-1]
            filters.append(self._resolve_filter(args, value, oper))

        return filters

    def _raw_get(self,
                 collection: str = 'default',
                 limit_: int = 50,
                 skip_: int = 0,
                 sort: Union[List, str] = ['timestamp__desc'],
                 group_id: Union[List, str] = [],
                 group_fields: Union[List, str] = [],
                 tzinfo: timezone = None,
                 **kwargs) -> List[BaseModel]:

        if group_id or group_fields or tzinfo:
            raise NotImplementedError(f"Not implemented yet in fake version. Try removing group_ and tzinfo fields")

        result = self._find(collection, filters=self._parse_filters(**kwargs))

        if limit_ == -1:
            return result[skip_:]

        return result[skip_:skip_ + limit_]

    @validate_resource_type
    def _get(self,
             resource_type: Type,
             collection: str = 'default',
             limit_: int = 50,
             skip_: int = 0,
             sort: Union[List, str] = ['timestamp__desc'],
             group_id: Union[List, str] = [],
             group_fields: Union[List, str] = [],
             tzinfo: timezone = None,
             **kwargs) -> List[BaseModel]:

        if group_id or group_fields or tzinfo:
            raise NotImplementedError(f"Not implemented yet in fake version. Try removing group_ and tzinfo fields")

        result = [resource_type(**v) for v in self._find(collection, filters=self._parse_filters(**kwargs))]

        if limit_ == -1:
            return result[skip_:]

        return result[skip_:skip_ + limit_]

    def raw_save(self, instances: List[Dict], collection: str = "default") -> List[Dict]:
        return self._write_to_collection(collection, instances)

    def save(self, instances: List[BaseModel], collection: str = "default") -> List[BaseModel]:
        data = [instance.dict() for instance in instances]
        self.raw_save(data, collection)
        return instances

    def get_db_size_gb(self) -> float:
        return 0.555555

    def get_dataframe(self, *args, **kwargs) -> VariableDataFrame:
        logger.info(f"[FAKED] getting dataframe {args}, {kwargs}")
        _data = self.raw_get(*args, **kwargs)
        _data = pd.DataFrame(_data)
        _data.set_index('timestamp', inplace=True)
        return _data

    def get_dataset(self, queries: List[Dict]) -> pd.DataFrame:
        dfs = [self.get_dataframe(**query) for query in queries]
        return pd.concat(dfs, axis=1)

    def raw_aggregate(self, collection: str, pipeline: List[Dict]) -> List[Dict]:
        raise NotImplementedError

    def save_dataframe(self, dataframe: pd.DataFrame, collection: str = 'default') -> None:
        logger.info(f"[FAKED] saving dataframe {dataframe.columns}")
        data = dataframe.reset_index().to_dict(orient="records")
        self.raw_save(instances=data, collection=collection)

    def create_index(self, collection: str, index: list) -> None:
        pass

    def get_output(self, query: Query) -> List[Dict]:
        return self.raw_get(
            collection=query.collection,
            limit_=query.limit,
            skip_=query.skip,
            sort=query.sort,
            add_fields=query.add_fields,
            group_id=query.group_id,
            group_fields=query.group_fields,
            rename_fields=query.rename_fields,
            project_fields=query.project_fields,
            tzinfo=timezone(timedelta(hours=query.timezone_offset)),
            **query.filters
        )


from furl import furl
from pydantic import BaseModel
from typing import Dict, Optional, Type, List
from requests import Session
from splight_abstract.storage import AbstractStorageClient
from remote_splight_lib.auth.auth import SplightAuthToken
from remote_splight_lib.settings import settings
from remote_splight_lib.exceptions import InvalidModel
from remote_splight_lib.storage.classmap import CLASSMAP
from splight_abstract.remote import AbstractRemoteClient


class StorageClient(AbstractStorageClient, AbstractRemoteClient):
    """Splight API Storage Client.
    Responsible for interacting with storage resources using HTTP requests
    to the Splight API.
    """

    def __init__(
        self,
        namespace: str = "default"
    ):
        """Class constructor

        Parameters
        ----------
        base_url : Union[str, furl]
            The URL for the Slight API
        token : Optional[SplightAuthToken]
            Instance of SplightAuthToken
        """
        super(StorageClient, self).__init__(namespace=namespace)
        self._base_url = furl(settings.SPLIGHT_PLATFORM_API_HOST)
        token = SplightAuthToken(
            access_key=settings.SPLIGHT_ACCESS_ID,
            secret_key=settings.SPLIGHT_SECRET_KEY,
        )
        self._session = Session()
        self._session.headers.update(token.header)

    @staticmethod
    def _get_model_data(constructor: Type):
        model_data = CLASSMAP.get(constructor)
        if not model_data:
            raise InvalidModel(constructor.schema()["title"])
        return model_data

    def _list(self, path: str, **kwargs):
        url = self._base_url / f"{path}/"
        params = self._parse_params(**kwargs)
        response = self._session.get(url, params=params)
        response.raise_for_status()
        return response.json()

    def _create(self, path: str, data: BaseModel) -> Dict:
        url = self._base_url / f"{path}/"
        response = self._session.post(url, json=data.dict())
        response.raise_for_status()
        return response.json()

    def _update(self, path: str, resource_id: str, data: BaseModel) -> Dict:
        url = self._base_url / f"{path}/{resource_id}/"
        response = self._session.put(url, json=data.dict())
        response.raise_for_status()
        return response.json()

    def _get(self, resource_type: Type, first=False, prefix: Optional[str] = None, limit_: int = -1, skip_: int = 0, **kwargs) -> List[BaseModel]:
        model_data = self._get_model_data(resource_type)
        path = model_data["path"]
        response = self._list(
            path,
            first=first,
            limit_=limit_,
            skip_=skip_,
            **kwargs,
        )
        parsed = [
            resource_type.parse_obj(resource)
            for resource in response["results"]
        ]
        return parsed

    def save(self, instance: BaseModel) -> BaseModel:
        raise NotImplementedError

    def copy(self, old_name: str, new_name: str):
        raise NotImplementedError

    def delete(self, resource_type: Type, id: str):
        raise NotImplementedError

    def download(self, resource_type: Type, id: str, target: str) -> str:
        model_data = self._get_model_data(resource_type)
        path = model_data["path"]
        url = self._base_url / f"{path}/{id}/download"
        response = self._session.get(url)
        response.raise_for_status()
        with open(target, "wb") as f:
            f.write(response.content)
        return response.content

    def get_temporary_link(self, filename: str, target: str) -> str:
        raise NotImplementedError



import sys
import time
import pandas as pd
import splight_models as spmodels
from functools import partial
from abc import abstractmethod
from tempfile import NamedTemporaryFile
from typing import Optional, Type, List, Dict, Tuple, Set, Any, Callable
from mergedeep import merge, Strategy as mergeStrategy
from collections import defaultdict
from pydantic import BaseModel
from functools import cached_property
from splight_lib.execution import ExecutionClient, Thread
from splight_lib.logging import logging
from splight_lib.settings import setup as default_setup
from splight_lib.shortcut import save_file as _save_file  # TODO unify storage with database
from splight_models import (
    CustomType,
    Deployment,
    Notification,
    StorageFile,
    ComponentDatalakeModel,
    Component,
    Command,
    Binding,
    ComponentObject,
)
from splight_models.component import EventNames, ComponentCommandUpdateEvent, ComponentCommandTriggerEvent, CommunicationEvent
from splight_models.component import ComponentCommand, ComponentCommandStatus
from splight_models.component import DATABASE_TYPES, NATIVE_TYPES, STORAGE_TYPES, Parameter


logger = logging.getLogger()


class RunnableMixin:
    healthcheck_interval = 5
    _HEALTH_FILE_PREFIX = "healthy_"
    _STARTUP_FILE_PREFIX = "ready_"

    def __init__(self):
        self.health_file = NamedTemporaryFile(prefix=self._HEALTH_FILE_PREFIX)
        self.startup_file = NamedTemporaryFile(
            prefix=self._STARTUP_FILE_PREFIX
        )
        self.execution_client.start(Thread(self.healthcheck))

    def healthcheck(self) -> None:
        self.terminated = False
        while not self.terminated:
            if not self.execution_client.healthcheck():
                logger.error("A task has failed")
                self.health_file.close()
                logger.error(f"Healthcheck file removed: {self.health_file}")
                sys.exit()
            time.sleep(self.healthcheck_interval)

    @abstractmethod
    def start(self):
        pass

    def terminate(self):
        self.terminated = True


class IndexMixin:
    def _load_client_indexes(self) -> None:
        indexes: List[Tuple[str, int]] = self.__get_indexes()
        logger.debug(f"Adding indexes: {indexes}")
        self.datalake_client.create_index(
            self.collection_name,
            indexes
        )

    def __get_indexes(self) -> List[Tuple[str, int]]:
        # order in indexes matters
        indexes: List[Tuple[str, int]] = [
            ("output_format", 1),
        ]
        unique_indexes: Set[str] = set()

        for output in self._spec.output:
            for field in output.fields:
                if field.filterable:
                    unique_indexes.add(field.name)

        for index in unique_indexes:
            indexes.append((index, 1))

        indexes.append(("timestamp", -1))

        return indexes


class HooksMixin:
    def _load_client_hooks(self):
        # Datalake
        self.datalake_client.add_pre_hook("save", self.__hook_insert_origin_save)
        self.datalake_client.add_pre_hook("save", self.__hook_lock_save_collection)
        self.datalake_client.add_pre_hook("save_dataframe", self.__hook_insert_origin_save_dataframe)
        self.datalake_client.add_pre_hook("save_dataframe", self.__hook_lock_save_collection)
        # Database
        self.database_client.add_pre_hook("save", self.__hook_transform_from_custom_instances)
        self.database_client.add_pre_hook("get", self.__hook_transform_from_custom_resource_type)
        self.database_client.add_pre_hook("delete", self.__hook_transform_from_custom_resource_type)
        self.database_client.add_pre_hook("count", self.__hook_transform_from_custom_resource_type)
        self.database_client.add_post_hook("get", self.__hook_transform_to_custom_instances)

    def __hook_insert_origin_save(self, *args, **kwargs):
        instances = kwargs.get("instances", [])
        for instance in instances:
            if not isinstance(instance, ComponentDatalakeModel):
                continue
            instance.instance_id = self.instance_id
            instance.instance_type = self.managed_class.__name__

        return args, kwargs

    def __hook_insert_origin_save_dataframe(self, *args, **kwargs):
        dataframe = kwargs.get("dataframe")
        dataframe["instance_id"] = self.instance_id
        dataframe["instance_type"] = self.managed_class.__name__
        kwargs["dataframe"] = dataframe
        return args, kwargs

    def __hook_lock_save_collection(self, *args, **kwargs):
        kwargs["collection"] = self.collection_name
        return args, kwargs

    def __hook_transform_from_custom_resource_type(self, *args, **kwargs):
        resource_type = kwargs["resource_type"]
        if resource_type and hasattr(self.custom_types, resource_type.__name__):
            kwargs["resource_type"] = ComponentObject
            kwargs["component_id"] = self.instance_id
            kwargs["type"] = resource_type.__name__
        return args, kwargs

    def __hook_transform_from_custom_instances(self, *args, **kwargs):
        instance = kwargs["instance"]
        if getattr(self.custom_types, type(instance).__name__, None):
            parsed_instance = ComponentObject(
                component_id=self.instance_id,
                type=type(instance).__name__,
                **self.unparse_parameters(instance)
            )
            kwargs["instance"] = parsed_instance
        return args, kwargs

    def __hook_transform_to_custom_instances(self, result: List[BaseModel]):
        parsed_result = []
        for object in result:
            if isinstance(object, ComponentObject):
                custom_object_data: List[Parameter] = object.data
                custom_object_data.extend([
                    Parameter(name=key, value=getattr(object, key))
                    for key in CustomType._reserved_names
                ])
                custom_object_model = getattr(self.custom_types, object.type)
                parsed_component_object = self.parse_parameters(custom_object_data)
                object = custom_object_model(**parsed_component_object)
            parsed_result.append(object)
        return parsed_result


class UtilsMixin:
    def get_history(self, **kwargs) -> pd.DataFrame:
        # TODO handle this with hooks?
        return self.datalake_client.get_dataframe(collection="default", **kwargs)

    def get_results(self, component: Component, output_model: ComponentDatalakeModel, **kwargs) -> pd.DataFrame:
        # TODO handle this with hooks?
        if output_model != getattr(component.output_model, output_model.__name__):
            raise ValueError(
                f"Output model {output_model.__name__} does not match component output"
            )

        return self.datalake_client.get_dataframe(
            collection=component.collection,
            output_format=output_model.__name__,
            **kwargs
        )

    def save_results(self, output_model: ComponentDatalakeModel, dataframe: pd.DataFrame) -> None:
        # TODO handle this with hooks?
        if output_model != getattr(self.output, output_model.__name__):
            raise ValueError(
                f"Output model {output_model.__name__} is not defined in the output"
            )

        try:
            for _, row in dataframe.iterrows():
                output_model.parse_obj(row.to_dict())
        except Exception:
            raise ValueError(f"Invalid dataframe: does not match output format")

        dataframe["output_format"] = output_model.__name__

        self.datalake_client.save_dataframe(
            dataframe=dataframe,
            collection=self.collection_name
        )

    def save_file(
        self,
        filename: str,
        prefix: Optional[str],
        asset_id: Optional[str],
        attribute_id: Optional[str],
        path: str,
        args: Dict,
    ) -> StorageFile:
        # TODO deprecate this
        return _save_file(
            self.storage_client,
            self.datalake_client,
            filename,
            prefix,
            asset_id,
            attribute_id,
            path,
            args,
        )

    def notify(self, notification: Notification):
        # TODO deprecate this.
        return self.database_client.save(notification)


class BindingsMixin:
    def _load_client_bindings(self):
        # Commands
        self.communication_client.bind(EventNames.COMPONENT_COMMAND_TRIGGER, self.__handle_component_command_trigger)

        # Bindings
        for binding in self.bindings:
            binding_function = getattr(self, binding.name)
            binding_model = getattr(spmodels, binding.object_type, ComponentObject)
            binding_event_name = binding_model.get_event_name(
                binding.object_type,
                binding.object_action
            )
            binding_handler = self.__handle_component_object_trigger if binding_model == ComponentObject else self.__handle_native_object_trigger
            self.communication_client.bind(
                binding_event_name,
                partial(
                    binding_handler,
                    binding_function,
                    binding.object_type
                )
            )
            logger.info(f"Binded event: {binding_event_name}")

    def __handle_native_object_trigger(self, binding_function: Callable, binding_object_type: str, data: str):
        assert self.bindings, "Please define .bindings to start."
        object_event = CommunicationEvent.parse_raw(data)
        object_model = getattr(spmodels, binding_object_type)
        binding_kwargs = object_model(**object_event.data)
        binding_function(binding_kwargs)

    def __handle_component_object_trigger(self, binding_function: Callable, binding_object_type: str, data: str):
        assert self.bindings, "Please define .bindings to start."
        component_object_event = CommunicationEvent.parse_raw(data)
        component_object: ComponentObject = ComponentObject(**component_object_event.data)
        custom_object_data = component_object.data
        custom_object_data.extend([
            Parameter(name=key, value=getattr(component_object, key))
            for key in CustomType._reserved_names
        ])
        custom_object_model = getattr(self.custom_types, binding_object_type)
        parsed_component_object = self.parse_parameters(custom_object_data)
        binding_kwargs = custom_object_model(**parsed_component_object)
        binding_function(binding_kwargs)

    def __handle_component_command_trigger(self, data: str):
        assert self.commands, "Please define .commands to start accepting request."
        component_command_event = ComponentCommandTriggerEvent.parse_raw(data)
        component_command: ComponentCommand = component_command_event.data
        command: Command = component_command.command
        try:
            command_function = getattr(self, command.name)
            command_kwargs_model = getattr(self.commands, command.name)
            parsed_command_kwargs = self.parse_parameters(command.dict()["fields"])
            command_kwargs = command_kwargs_model(**parsed_command_kwargs)
            # .dict is not keeping the models of subkeys
            command_kwargs = {
                str(field): getattr(command_kwargs, str(field)) for field in command_kwargs.__fields__
            }
            component_command.response.return_value = str(command_function(**command_kwargs))
            component_command.status = ComponentCommandStatus.SUCCESS
        except Exception as e:
            component_command.response.error_detail = str(e)
            component_command.status = ComponentCommandStatus.ERROR
        component_command_callback_event = ComponentCommandUpdateEvent(data=component_command)
        self.communication_client.trigger(component_command_callback_event)


class ParametersMixin:
    def unparse_parameters(self, instance: Dict) -> List[Dict]:
        custom_type = getattr(self.custom_types, type(instance).__name__, None)
        if custom_type is None:
            raise NotImplementedError
        reserved_parameters = {k: v for k, v in instance.dict().items() if k in CustomType._reserved_names}
        custom_parameters = {k: v for k, v in instance.dict().items() if k not in CustomType._reserved_names}
        fields = []
        for key, obj in custom_parameters.items():
            field = getattr(custom_type.Meta, key)
            if field is None:
                continue
            value = obj

            if field.type not in NATIVE_TYPES:
                value = [o["id"] for o in obj] if field.multiple else obj["id"]

            fields.append(
                Parameter(
                    name=field.name,
                    value=value,
                    type=field.type,
                    multiple=field.multiple,
                    required=field.required,
                    choices=field.choices,
                    depends_on=field.depends_on,
                    sensitive=field.sensitive,
                )
            )
        return {"data": fields, **reserved_parameters}

    def parse_parameters(self, parameters: List[Dict]) -> Dict:
        parameters = self._fetch_and_reload_component_objects_parameters(parameters)
        object_ids: Dict[str, Dict[str, List]] = self._get_object_ids(parameters)
        objects: Dict[str, BaseModel] = self._fetch_objects(object_ids)
        parameters = self._reload_parameters(parameters, objects=objects)
        transformed_parameters = self._transform_parameters(parameters)
        return transformed_parameters

    def _fetch_and_reload_component_objects_parameters(self, parameters: List[Dict]) -> List[Dict]:
        reloaded_parameters = []
        for raw_parameter in parameters:
            raw_parameter = raw_parameter.dict() if not isinstance(raw_parameter, dict) else raw_parameter
            parameter = raw_parameter.copy()
            type = parameter["type"]
            value = parameter["value"]
            multiple = parameter["multiple"]
            if type in NATIVE_TYPES or type in DATABASE_TYPES or type in STORAGE_TYPES:
                parameter["value"] = value
            else:
                object_ids = value if multiple else [value]
                objects = self.database_client.get(ComponentObject, id__in=object_ids)
                for o in objects:
                    component_object_data = [
                        Parameter(name=key, value=getattr(o, key))
                        for key in CustomType._reserved_names
                    ]
                    o.data.extend(component_object_data)
                objects = [o.data for o in objects]
                parameter["value"] = [self._fetch_and_reload_component_objects_parameters(o) for o in objects] if multiple else self._fetch_and_reload_component_objects_parameters(objects[0])
            reloaded_parameters.append(parameter)
        return reloaded_parameters

    def _get_object_ids(self, parameters: List[Dict]) -> Dict[str, Dict[str, List]]:
        ids = {
            "database": defaultdict(list),
            "storage": defaultdict(list),
        }
        for parameter in parameters:
            parameter = parameter.dict() if not isinstance(parameter, dict) else parameter
            values = parameter["value"] if parameter["multiple"] else [parameter["value"]]
            if parameter["type"] in NATIVE_TYPES:
                continue
            elif parameter["type"] in DATABASE_TYPES:
                for value in values:
                    ids["database"][parameter["type"]].append(value)
            elif parameter["type"] in STORAGE_TYPES:
                for value in values:
                    ids["storage"][parameter["type"]].append(value)
            else:
                for value in values:
                    ids = merge(ids, self._get_object_ids(value), strategy=mergeStrategy.ADDITIVE)
        return ids

    def _fetch_objects(self, ids_to_fetch: Dict) -> Dict[str, BaseModel]:
        res: Dict = {
            None: None
        }
        for type, ids_ in ids_to_fetch["database"].items():
            objs = self.database_client.get(DATABASE_TYPES[type], id__in=ids_)
            res.update({obj.id: obj for obj in objs})

        for type, ids_ in ids_to_fetch["storage"].items():
            objs = self.storage_client.get(STORAGE_TYPES[type], id__in=ids_)
            res.update({obj.id: obj for obj in objs})
        return res

    def _reload_parameters(self, parameters: List[Dict], objects: Dict) -> List[Dict]:
        reloaded_parameters = []
        for raw_parameter in parameters:
            raw_parameter = raw_parameter.dict() if not isinstance(raw_parameter, dict) else raw_parameter
            parameter = raw_parameter.copy()
            type = parameter["type"]
            value = parameter["value"]
            multiple = parameter["multiple"]
            if type in NATIVE_TYPES:
                parameter["value"] = value
            elif type in DATABASE_TYPES or type in STORAGE_TYPES:
                parameter["value"] = [objects[val] for val in value] if multiple else objects[value]
            else:
                parameter["value"] = [self._reload_parameters(value, objects) for value in value] if multiple else self._reload_parameters(value, objects)
            reloaded_parameters.append(parameter)
        return reloaded_parameters

    def _transform_parameters(self, parameters: List[Dict]) -> Dict:
        parameters_dict: Dict = {}
        for parameter in parameters:
            type = parameter["type"]
            name = parameter["name"]
            value = parameter["value"]
            multiple = parameter["multiple"]

            if (value is [] or value == '') and type != "str":
                value = None

            if type in NATIVE_TYPES:
                parameters_dict[name] = value
            elif type in DATABASE_TYPES or type in STORAGE_TYPES:
                parameters_dict[name] = value
            else:
                parameters_dict[name] = [self._transform_parameters(val) for val in value] if multiple else self._transform_parameters(value)
        return parameters_dict


class AbstractComponent(RunnableMixin, HooksMixin, UtilsMixin, IndexMixin, BindingsMixin, ParametersMixin):
    collection_name = "default"
    managed_class: Type = Component
    database_client_kwargs: Dict[str, Any] = {}
    datalake_client_kwargs: Dict[str, Any] = {}
    deployment_client_kwargs: Dict[str, Any] = {}
    storage_client_kwargs: Dict[str, Any] = {}
    communication_client_kwargs: Dict[str, Any] = {}

    def __init__(self, run_spec: dict, initial_setup: Optional[dict] = None, *args, **kwargs):
        self._spec: Deployment = Deployment(**run_spec)
        self._setup = default_setup
        if initial_setup:
            self._setup.configure(initial_setup)

        self.version: str = self._spec.version
        self.namespace = self._setup.settings.NAMESPACE
        self.instance_id = self._setup.settings.COMPONENT_ID
        self.collection_name = str(self.instance_id)

        self._load_instance_data()
        self._load_clients()
        self._load_spec_models()
        self._load_input_model()
        self._load_client_hooks()
        self._load_client_indexes()
        self._load_client_bindings()
        super().__init__(*args, **kwargs)  # This is calling RunnableMixin.init() only

    @property
    def spec(self) -> Deployment:
        return self._spec

    @property
    def setup(self):
        return self._setup

    @cached_property
    def instance(self):
        return self.database_client.get(
            resource_type=self.managed_class, id=self.instance_id, first=True
        )

    def _load_instance_data(self):
        self.collection_name = str(self.instance_id)
        self.communication_client_kwargs['instance_id'] = self.instance_id

    def _load_spec_models(self):
        self.output: Type = self._spec.output_model
        self.custom_types: Type = self._spec.custom_types_model
        self.commands: Type = self._spec.commands_model
        self.bindings: List[Binding] = self._spec.bindings

    def _load_input_model(self):
        raw_spec = self.spec.dict()
        parsed_input_parameters = self.parse_parameters(raw_spec["input"])
        self.input: BaseModel = self._spec.input_model(**parsed_input_parameters)

    def _load_clients(self):
        self.database_client = self.setup.DATABASE_CLIENT(
            namespace=self.namespace,
            **self.database_client_kwargs
        )
        self.datalake_client = self.setup.DATALAKE_CLIENT(
            namespace=self.namespace,
            **self.datalake_client_kwargs
        )
        self.deployment_client = self.setup.DEPLOYMENT_CLIENT(
            namespace=self.namespace,
            **self.deployment_client_kwargs
        )
        self.storage_client = self.setup.STORAGE_CLIENT(
            namespace=self.namespace,
            **self.storage_client_kwargs
        )
        self.communication_client = self.setup.COMMUNICATION_CLIENT(
            namespace=self.namespace,
            **self.communication_client_kwargs
        )
        self.execution_client = ExecutionClient(namespace=self.namespace)
        self.blockchain_client = self.setup.BLOCKCHAIN_CLIENT(namespace=self.namespace)



import os
from typing import Optional
from splight_models import *
from splight_lib import logging
from splight_lib.settings import setup


COLLECTION = 'files'
DatalakeClient = setup.DATALAKE_CLIENT
StorageClient = setup.STORAGE_CLIENT
logger = logging.getLogger()


def save_file(storage_client: StorageClient, datalake_client: DatalakeClient,
              filename, prefix: Optional[str] = None,
              asset_id: Optional[str] = None, attribute_id: Optional[str] = None, path: str = '', args: Dict = None) -> StorageFile:
    if not os.path.exists(filename):
        logger.error("File must be locally found")
        raise Exception()
    args = args if args else dict()
    if args.get('file_id',None) is not None:
        msg = "Can't overwrite file_id"
        logger.error(msg)
        raise Exception(msg)
    file = storage_client.save(StorageFile(file=filename,content_type='str'), prefix=prefix)
    args['file_id'] = file.id
    var = Variable(asset_id=asset_id, attribute_id=attribute_id, path=path, args=args)
    datalake_client.save(Variable, instances=[var], collection=COLLECTION)
    return file


import random
from hexbytes import HexBytes

from splight_abstract import AbstractBlockchainClient, FunctionTransactError
from splight_models import CallResponse, Transaction


def get_random_hex(size: int = 256) -> HexBytes:
    return HexBytes(random.getrandbits(size))


class FakeBlockchainClient(AbstractBlockchainClient):

    def __init__(self, *args, **kwargs):
        pass

    def get_balance(self) -> int:
        return 111111111111111111

    def call(self, method: str, *args) -> CallResponse:
        return CallResponse(
            name=method,
            value=random.randint(0, 10000)
        )

    def transact(self, method: str, *args, **kwargs) -> Transaction:
        if method not in ["mint", "burn", "transfer"]:
            raise FunctionTransactError(method)

        return Transaction.parse_obj({
            "blockHash": get_random_hex(size=256),
            "blockNumber": 102812,
            "contractAddress": None,
            "cumulativeGasUsed": 40662,
            "from": get_random_hex(size=256),
            "gasUsed": 40662,
            "isPrivacyMarkerTransaction": False,
            "logsBloom": get_random_hex(size=1024),
            "status": 1,
            "to": get_random_hex(size=256),
            "transactionHash": get_random_hex(size=256),
            "transactionIndex": 0,
            "type": "0x0"
        })

    def get_transaction(self, tx_hash: HexBytes) -> Transaction:

        return Transaction.parse_obj({
            "blockHash": get_random_hex(size=256),
            "blockNumber": 102812,
            "contractAddress": None,
            "cumulativeGasUsed": 40662,
            "from": get_random_hex(size=256),
            "gasUsed": 40662,
            "isPrivacyMarkerTransaction": False,
            "logsBloom": get_random_hex(size=1024),
            "status": 1,
            "to": get_random_hex(size=256),
            "transactionHash": tx_hash,
            "transactionIndex": 0,
            "type": "0x0"
        })



from pydantic import BaseSettings


class SplightRemoteSettings(BaseSettings):
    SPLIGHT_PLATFORM_API_HOST: str = "https://api.splight-ae.com"
    SPLIGHT_ACCESS_ID: str
    SPLIGHT_SECRET_KEY: str

settings = SplightRemoteSettings()



from .auth import AbstractAuthClient, AuthenticationError
from splight_abstract.notification import *
from splight_abstract.cache import *
from splight_abstract.communication import *
from splight_abstract.database import *
from splight_abstract.datalake import *
from splight_abstract.deployment import *
from splight_abstract.hub import *
from splight_abstract.client import *
from splight_abstract.storage import *
from splight_abstract.blockchain import *
from splight_abstract.remote import *



from unittest import TestCase


class TestShortcut(TestCase):
    def test_import_shortcut(self):
        from splight_lib.shortcut import (
            notify,
            save_file,
            OrganizationHandler,
        )


import json
from abc import abstractmethod
from pydantic import BaseModel
from typing import Dict, Type, List, Optional
from splight_abstract.client import AbstractClient, QuerySet
from splight_models.deployment import DeploymentEvent


class AbstractDeploymentClient(AbstractClient):
    @abstractmethod
    def save(self, instance: BaseModel) -> BaseModel:
        pass

    def get(self, *args, **kwargs) -> QuerySet:
        return QuerySet(self, *args, **kwargs)

    @abstractmethod
    def _get(self, resource_type: Type, id: str = '', first=False, limit_: int = -1, skip_: int = 0, **kwargs) -> List[BaseModel]:
        pass

    @abstractmethod
    def delete(self, resource_type: Type, id: str) -> None:
        pass

    def get_deployment_logs(self, id: str, limit: Optional[int] = None, since: Optional[str] = None, previous: bool = False) -> List[str]:
        pass

    @classmethod
    @abstractmethod
    def verify_header(cls, payload: bytes, signature: str) -> None:
        pass

    @classmethod
    def construct_event(cls, payload: bytes, signature: str) -> DeploymentEvent:
        cls.verify_header(payload, signature)
        event = json.loads(payload)
        return DeploymentEvent.parse_obj(event)



from splight_models.base import SplightBaseModel
from pydantic import validator
from typing import Dict


def namespace_transform(raw: str) -> str:
    return raw.lower().replace("_", "")


class Namespace(SplightBaseModel):
    id: str
    environment: Dict = {}
    _extract_id = validator('id', pre=True, allow_reuse=True)(namespace_transform)




class InvalidModel(Exception):
    def __init__(self, model_name: str):
        self._msg = f"Model {model_name} is not a valid database model"

    def __str__(self) -> str:
        return self._msg



class InvalidReference(Exception):
    pass


class CyclicReference(Exception):
    pass


class DuplicatedAttribute(Exception):
    pass


class CrossNamespaceTryException(Exception):
    def __init__(self, namespace, conflicts, *args, **kwargs) -> None:
        self.namespace = namespace
        self.conflicts = conflicts
        message = f"Found conflicts in fields: {conflicts}. Current tenant: {self.tenant}"
        super().__init__(message, *args, **kwargs)


class LockedGraphException(Exception):
    pass


class CrossGraphException(Exception):
    pass



from typing import Any, Dict, Optional

from requests import Session
from simplejson.errors import JSONDecodeError


class BaseEndpoint:
    def __init__(self, base_url: str, session: Session):
        self._session = session
        self._base_url = base_url
        self._url = f"{self._base_url.strip('/')}/{self.PATH.strip('/')}/"

    def _request_content(self, response: Any):
        try:
            return response.json() if response.content else None
        except JSONDecodeError as e:
            raise Exception(str(response.content))

    def _get_request(self, url: str, params: Optional[Dict] = None):
        response = self._session.get(url, params=params)
        content = self._request_content(response)
        return content, response.status_code

    def _post_request(self, url: str, data: Dict):
        response = self._session.post(url, json=data)
        content = self._request_content(response)
        return content, response.status_code

    def _put_request(self, url: str, data: Dict):
        response = self._session.put(url, json=data)
        content = self._request_content(response)
        return content, response.status_code

    def _patch_request(self, url: str, data: Dict):
        response = self._session.patch(url, json=data)
        content = self._request_content(response)
        return content, response.status_code

    def _delete_request(self, url: str):
        response = self._session.delete(url)
        content = self._request_content(response)
        return content, response.status_code


class ListMixin:
    def list(self, params: Optional[Dict] = None):
        return self._get_request(url=self._url, params=params)


class RetrieveMixin:
    def retrieve(self, resource_id: str):
        url = f"{self._url}{resource_id}/"
        return self._get_request(url=url)


class CreateMixin:
    def create(self, data: Dict[str, Any]):
        return self._post_request(url=self._url, data=data)


class UpdateMixin:
    def update(self, resource_id: str, data: Dict[str, Any]):
        url = f"{self._url}{resource_id}/"
        return self._put_request(url=url, data=data)


class PartialUpdateMixin:
    def partial_update(self, resource_id: str, data: Dict[str, Any]):
        url = f"{self._url}{resource_id}/"
        return self._patch_request(url=url, data=data)


class DestroyMixin:
    def destroy(self, resource_id: str):
        url = f"{self._url}{resource_id}/"
        return self._delete_request(url=url)



from enum import Enum
from typing import Dict, Union, List, Optional
from pydantic import validator, root_validator
from splight_models.base import SplightBaseModel

class QuerySource(str, Enum):
    # TODO: remove this once component types is unified
    COMPONENT = "Component"
    ALGORITHM = "Algorithm"
    CONNECTOR = "Connector"

class Query(SplightBaseModel):
    # TODO: Update query after all component types are unified
    id: Optional[str]
    name: str
    description: Optional[str]
    source: QuerySource
    component_id: str = None
    output_format: str
    target: str = None
    collection: str = None
    filters: Dict = {}
    limit: int = 10000
    skip: int = 0
    sort: Union[List, str] = ['timestamp__desc']
    add_fields: Union[List, str] = []
    group_id: Union[List, str] = []
    group_fields: Union[List, str] = []
    rename_fields: Union[List, str] = []
    project_fields: Union[List, str] = []
    timezone_offset: float = 0.0 # UTC
    query_params: str = None

    @root_validator(pre=True)
    def component_id_validate(cls, values):
        component_id = values.get("component_id")
        if component_id:
            return values
        source = values.get("source")
        if source == QuerySource.CONNECTOR:
            values["component_id"] = "default"
            return values
        raise ValueError(f"component_id is required when source is {QuerySource.ALGORITHM}")

    @validator("filters", always=True)
    def filters_validate(cls, filters, values):
        source = values.get("source")
        component_id = values.get("component_id")
        if source == QuerySource.CONNECTOR and component_id != "default":
            filters["instance_id"] = component_id
        return filters

    @validator("collection", always=True)
    def collection_validate(cls, collection, values):
        source = values.get("source")
        if source == QuerySource.ALGORITHM:
            return values.get("component_id")
        return "default"

    @validator("target", always=True)
    def target_validate(cls, target, values):
        source = values.get("source")
        if source == QuerySource.CONNECTOR:
            target = "value"
        if not target:
            raise ValueError("Target is required for algorithm outputs")
        return target

    @validator("query_params", always=True)
    def query_params_validate(cls, query_params, values):
        if query_params is None:
            query_params = f"?source={values['collection']}&"
            query_params += "&".join([f"{k}={','.join([str(i) for i in v])}" if type(v) == list else f"{k}={v}" for k, v in values["filters"].items()]) + "&" if values["filters"] else ""
            query_params_fields = ["limit", "skip", "sort"]
            underscored_fields = ["limit", "skip"]
            for k in query_params_fields:
                field = k if not k in underscored_fields else k + "_"
                if type(values[k]) == list:
                    query_params += f"{field}={','.join(values[k])}&"
                else:
                    query_params += f"{field}={values[k]}&"
            query_params = query_params[:-1]
        return query_params



from typing import Dict, List, Optional
from splight_models.base import SplightBaseModel
from splight_models.constants import ComponentSize, ComponentType, DeploymentStatus, LogginLevel, RestartPolicy
from splight_models.component import BaseComponent


class Deployment(BaseComponent):
    # run-spec
    id: Optional[str] = None
    type: ComponentType = ComponentType.COMPONENT
    component_id: Optional[str] = None
    status: Optional[DeploymentStatus] = None
    status_conditions: Optional[List[Dict]] = None
    container_statuses: Optional[List[Dict]] = None
    tags: Optional[List[str]] = []

    # Template vars for deployment
    namespace: Optional[str] = None
    component_capacity: ComponentSize = ComponentSize.medium
    log_level: LogginLevel = LogginLevel.debug
    restart_policy: RestartPolicy = RestartPolicy.ALWAYS

    # CLI pre setup
    secrets: Optional[str] = None

    # TODO: Remove the following attributes
    access_id: Optional[str] = None
    secret_key: Optional[str] = None
    hub_api_host: Optional[str] = None
    api_host: Optional[str] = None

    class Config:
        use_enum_values = True

    @property
    def service_name(self):
        id = str(self.id).lower()
        type_id = str(self.type).lower()
        return f"service-{type_id}-{id}"

    @property
    def deployment_name(self):
        # TODO remove this. we are no longeer using deployments.
        id = str(self.id).lower()
        type_id = str(self.type).lower()
        return f"deployment-{type_id}-{id}"


class DeploymentEvent(SplightBaseModel):
    severity: str
    reason: str
    message: str
    namespace: str
    object_id: str
    object_type: str



from typing import Optional
from pydantic import validator
from splight_models.component import BaseComponent
from typing import List

VERIFICATION_CHOICES = ['verified', 'unverified', 'official']


class HubComponent(BaseComponent):
    id: Optional[str]
    name: str
    type: str = "Component"
    splight_cli_version: str
    build_status: Optional[str]
    description: Optional[str]
    privacy_policy: Optional[str] = None
    tenant: Optional[str] = None
    readme: Optional[str]
    picture: Optional[str]
    file: Optional[str]
    verification: Optional[str]
    created_at: Optional[str]
    last_modified: Optional[str]
    tags: List[str] = []

    @validator('verification', pre=True, always=True)
    def set_verification_now(cls, v):
        if v:
            assert v in VERIFICATION_CHOICES, 'Verification value not allowed.'
        return v


class HubComponentVersion(HubComponent):
    pass



from enum import Enum
from typing import Callable, Dict
from abc import abstractmethod, abstractproperty
from splight_abstract.client import AbstractClient
from splight_models import CommunicationContext, CommunicationEvent


class CommunicationClientStatus(str, Enum):
    STARTING = 'starting'
    READY = 'ready'
    FAILED = 'failed'
    ERROR =  'error'


class AbstractCommunicationClient(AbstractClient):
    def __init__(self, instance_id: str = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance_id = instance_id

    @abstractproperty
    def context(self) -> CommunicationContext:
        pass
    
    @abstractproperty
    def status(self):
        pass

    @abstractmethod
    def bind(self, event_name: str, event_handler: Callable) -> None:
        pass

    @abstractmethod
    def unbind(self, event_name: str, event_handler: Callable) -> None:
        pass

    @abstractmethod
    def trigger(self, event: CommunicationEvent):
        pass

    @abstractmethod
    def authenticate(self, channel_name: str, socket_id: str, custom_data: Dict = None) -> dict:
        pass



class TransactionError(Exception):
    def __init__(self, tx_hash: str):
        self._msg = f"An error occurred during transaction {tx_hash}"

    def __str__(self) -> str:
        return self._msg


class FunctionCallError(Exception):
    def __init__(self, name: str):
        self._msg = f"An error occured calling function {name}"

    def __str__(self) -> str:
        return self._msg


class ProviderConnectionError(Exception):
    def __init__(self):
        self._msg = "Unable to connect to provider"

    def __str__(self) -> str:
        return self._msg


class ContractNotLoaded(Exception):
    def __init__(self):
        self._msg = "Smart Contract not loaded yet"

    def __str__(self) -> str:
        return self._msg


class FunctionTransactError(Exception):
    def __init__(self, method: str):
        self._msg = f"Method {method} not defined in smart contract"

    def __str__(self) -> str:
        return self._msg



import six
import hmac
import time
from hashlib import sha256
from splight_lib.auth.exceptions import SignatureVerificationError


def utf8(value):
    if six.PY2 and isinstance(value, six.text_type):
        return value.encode("utf-8")
    else:
        return value


class HmacSignature(object):
    EXPECTED_SCHEME = "v1"
    DEFAULT_TOLERANCE = 300

    def __init__(self, secret="", *args, **kwargs):
        self.secret = secret
        super().__init__(*args, **kwargs)

    @staticmethod
    def _compute_signature(payload, secret):
        mac = hmac.new(
            secret.encode("utf-8"),
            msg=payload.encode("utf-8"),
            digestmod=sha256,
        )
        return mac.hexdigest()

    def compute_header_signature(self, payload: str):
        """
        Compute headers signature using hmac auth method.
        """
        timestamp = int(time.time())
        signed_payload = "%d.%s" % (timestamp, payload.decode("ascii"))
        mac_hexdigest = self._compute_signature(signed_payload, self.secret)
        return f"t={timestamp},{self.EXPECTED_SCHEME}={mac_hexdigest}"

    @staticmethod
    def _get_timestamp_and_signatures(header, scheme):
        list_items = [i.split("=", 2) for i in header.split(",")]
        timestamp = int([i[1] for i in list_items if i[0] == "t"][0])
        signatures = [i[1] for i in list_items if i[0] == scheme]
        return timestamp, signatures

    @classmethod
    def verify_header(cls, payload, header, secret, tolerance=DEFAULT_TOLERANCE):
        if hasattr(payload, "decode"):
            payload = payload.decode("utf-8")

        try:
            timestamp, signatures = cls._get_timestamp_and_signatures(
                header, cls.EXPECTED_SCHEME
            )
        except Exception:
            raise SignatureVerificationError(
                "Unable to extract timestamp and signatures from header",
                header,
                payload,
            )

        if not signatures:
            raise SignatureVerificationError(
                "No signatures found with expected scheme "
                "%s" % cls.EXPECTED_SCHEME,
                header,
                payload,
            )

        signed_payload = "%d.%s" % (timestamp, payload)
        expected_sig = cls._compute_signature(signed_payload, secret)
        if not any(hmac.compare_digest(utf8(expected_sig), utf8(s)) for s in signatures):
            raise SignatureVerificationError(
                "No signatures found matching the expected signature for "
                "payload",
                header,
                payload,
            )

        if tolerance and timestamp < time.time() - tolerance:
            raise SignatureVerificationError(
                "Timestamp outside the tolerance zone (%d)" % timestamp,
                header,
                payload,
            )

        return True




class SplightAuthToken:
    def __init__(self, access_key: str, secret_key: str):
        self._token = f"Splight {access_key} {secret_key}"

    @property
    def header(self):
        return {"Authorization": self._token}



# TODO remove this models from here.
import re
import json
from datetime import datetime, timezone
from typing import Optional, Union
from pydantic import validator, Json, Field, validator
from hexbytes import HexBytes
from splight_models.base import SplightBaseModel


class BlockchainContractSubscription(SplightBaseModel):
    id: Optional[str]
    asset_id: str
    attribute_id: str
    contract_id: str
    last_checkpoint: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))


class BlockchainContract(SplightBaseModel):
    id: Optional[str]
    name: str
    description: str = ""
    address: str
    abi_json: Json

    @validator('address', pre=True, always=True)
    def set_account_id_now(cls, v):
        regex = r"(\b0x[a-fA-F0-9]{40}\b)"
        assert re.match(regex, v), 'Account value not allowed it should be a hex str.'
        return v

    @validator('abi_json', pre=True, always=True)
    def set_abi_json_now(cls, v):
        if isinstance(v, list):
            return json.dumps(v)
        return v


class CallResponse(SplightBaseModel):
    name: str
    value: Union[int, float, str]


class Transaction(SplightBaseModel):
    from_account: str = Field(..., alias="from")
    to_account: Optional[str] = Field(None, alias="to")
    status: Optional[int] = 0
    contract_address: Optional[str] = Field(None, alias="contractAddress")
    block_hash: str = Field(..., alias="blockHash")
    block_number: int = Field(..., alias="blockNumber")
    transaction_hash: str = Field(..., alias="transactionHash")
    transaction_index: int = Field(..., alias="transactionIndex")

    @validator("from_account", "to_account", "block_hash", "transaction_hash", "contract_address", pre=True)
    def cast_to_str(cls, value):
        return value.hex() if isinstance(value, HexBytes) else value



from unittest import TestCase
from fake_splight_lib.cache import FakeCacheClient
from splight_abstract.cache import get_cache, flush_cache


CACHE_CLIENT = FakeCacheClient()


class TestCacheFunctions(TestCase):

    def setUp(self):
        CACHE_CLIENT.clear()
        self.called_count = 0

    @get_cache(CACHE_CLIENT, 'cached_dummy', 'param')
    def cached_dummy(self, param, key='p'):
        self.called_count += 1
        return {key: param}

    @flush_cache(CACHE_CLIENT, 'cached_dummy', 'param')
    def flush_dummy(self, param):
        pass

    @flush_cache(CACHE_CLIENT, 'cached_dummy', prefix_match=True)
    def flush_dummy_prefix(self):
        pass

    def test_get_cache_same_param(self):
        self.assertEqual(self.called_count, 0)
        self.cached_dummy(1)
        self.cached_dummy(1)
        self.assertEqual(self.called_count, 1)

    def test_get_cache_multiple_params(self):
        self.cached_dummy(1)
        self.cached_dummy(2)
        self.cached_dummy(2)
        self.assertEqual(self.called_count, 2)

    def test_flush_cache(self):
        self.cached_dummy(1)
        self.assertEqual(self.called_count, 1)
        self.flush_dummy(1)
        self.cached_dummy(1)
        self.assertEqual(self.called_count, 2)

    def test_cache_values(self):
        EXPECTED_RESULT = self.cached_dummy(1)
        res = self.cached_dummy(1)
        self.assertEqual(self.called_count, 1)
        self.assertEqual(res, EXPECTED_RESULT)

    def test_flush_cache_multiple_values(self):
        self.cached_dummy(1, 'hola')
        self.cached_dummy(2, 'manola')
        self.flush_dummy_prefix()



import logging
from enum import Enum, IntEnum


class SeverityType(str, Enum):
    system = 'system'
    info = 'info'
    low = 'low'
    medium = 'medium'
    high = 'high'
    critical = 'critical'


class ComponentSize(str, Enum):
    small = "small"
    medium = "medium"
    large = "large"
    very_large = "very_large"

    def __str__(self):
        return self.value


class LogginLevel(IntEnum):
    critical = logging.CRITICAL
    error = logging.ERROR
    warning = logging.WARNING
    info = logging.INFO
    debug = logging.DEBUG
    notset = logging.NOTSET

    def __str__(self):
        return str(self.value)

    @classmethod
    def _missing_(cls, value):
        return super()._missing_(int(value))


class RestartPolicy(str, Enum):
    ALWAYS = "Always"
    ON_FAILURE = "OnFailure"
    NEVER = "Never"


class ComponentType(str, Enum):
    # TODO: Remove this
    COMPONENT = "Component"
    ALGORITHM = "Algorithm"
    NETWORK = "Network"
    CONNECTOR = "Connector"
    SYSTEM = "System"


class DeploymentStatus(str, Enum):
    PENDING = "Pending"
    RUNNING = "Running"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    UNKNOWN = "Unknown"


class ComponentStatus(str, Enum):
    STOPPED = "Stopped"
    PENDING = "Pending"
    RUNNING = "Running"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    UNKNOWN = "Unknown"


class PrivacyPolicy(str, Enum):
    PUBLIC = "public"
    PRIVATE = "private"


class VerificationLevel(str, Enum):
    VERIFIED = "verified"
    UNVERIFIED = "unverified"
    OFFICIAL = "official"


class BuildStatus(str, Enum):
    PENDING = "pending"
    BUILDING = "building"
    FAILED = "failed"
    SUCCESS = "success"
    UNKNOWN = "unknown"



from typing import Dict, Any
from .datalake import DatalakeModel


class UserActivityType:
    CREATE = "CREATE"
    UPDATE = "UPDATE"
    DELETE = "DELETE"
    READ = "READ"


class UserActivity(DatalakeModel):
    action: str
    user: Dict[str, Any]
    object: Dict[str, Any]
    details: Dict[str, Any]



from setuptools import setup, find_packages


with open('requirements.txt') as fp:
    install_requires = fp.readlines()

dependency_links = [
    # External repoitories different from pypi
]

setup(
    name='splight-lib',
    version='0.5.12',
    author='Splight',
    author_email='factory@splight-ae.com',
    packages=find_packages(),
    package_data={},
    include_package_data=True,
    scripts=[],
    url=None,
    zip_safe=False,
    license='LICENSE.txt',
    description='Library for public use. Splight',
    long_description=open('README.md').read(),
    install_requires=install_requires,
    dependency_links=dependency_links
)



from datetime import datetime
from typing import Dict, List, Optional

from splight_models.base import SplightBaseModel


class Role(SplightBaseModel):
    id: str
    name: str
    description: Optional[str] = None
    permission: List = []


class UserMetadata(SplightBaseModel):
    language: str = "en"
    theme: str = "dark"
    picture_color: str = "light-green"


class AppMetadata(SplightBaseModel):
    is_manager: bool = False


class User(SplightBaseModel):
    user_id: str
    organization_id: Optional[str] = None
    username: Optional[str] = None
    name: str
    email: str
    picture: str
    permissions: Optional[List[str]] = None
    last_login: Optional[datetime] = None
    # picture_color: Optional[str] = None
    roles: Optional[List[Dict]] = None
    role: Optional[str] = None
    user_metadata: Optional[UserMetadata] = None
    app_metadata: Optional[AppMetadata] = None
    is_authenticated: bool = True
    blocked: Optional[bool] = None

    @property
    def is_anonymous(self):
        return False

    def has_any_permission(self, permissions: List[str]) -> bool:
        return any(set(self.permissions).intersection(permissions))

    def __str__(self) -> str:
        return self.email

    def __repr__(self) -> str:
        return f"<User {self.email}>"


class OrganizationMetadata(SplightBaseModel):
    billing_id: Optional[str] = None
    blockchain_id: Optional[str] = None
    manager_email: Optional[str] = None


class Organization(SplightBaseModel):
    id: str
    name: str
    display_name: str
    metadata: OrganizationMetadata = OrganizationMetadata()


class Inviter(SplightBaseModel):
    name: str


class Invitee(SplightBaseModel):
    email: str


class Invitation(SplightBaseModel):
    id: Optional[str] = None
    organization_id: Optional[str] = None
    inviter: Inviter
    invitee: Invitee
    app_metadata: AppMetadata
    user_metadata: UserMetadata
    invitation_url: str
    created_at: datetime
    expires_at: datetime
    client_id: str
    roles: List[str] = []
    ticket_id: str



import os
from typing import Optional, List
from .base import SplightBaseModel


class StorageFile(SplightBaseModel):
    id: Optional[str] = None
    file: str
    content_type: Optional[str] = None

    def __init__(self, **kwargs):
        file = kwargs.pop('file', None)
        file = file.replace('/', os.sep)
        file = file.replace('\\', os.sep)
        kwargs['file'] = file
        super(StorageFile, self).__init__(**kwargs)

    @property
    def name(self):
        return self.file.split(os.sep)[-1]



from abc import abstractmethod
from pydantic import BaseModel
from typing import Optional, Type, List
from splight_abstract.client import AbstractClient, QuerySet

class AbstractStorageClient(AbstractClient):

    @abstractmethod
    def save(self, instance: BaseModel) -> BaseModel:
        pass

    def get(self, *args, **kwargs) -> QuerySet:
        return QuerySet(self, *args, **kwargs)

    @abstractmethod
    def _get(self, resource_type: Type, first=False, prefix: Optional[str] = None, limit_: int = -1, skip_: int = 0, **kwargs) -> List[BaseModel]:
        pass

    @abstractmethod
    def copy(self, old_name: str, new_name: str):
        pass

    @abstractmethod
    def delete(self, resource_type: Type, id: str) -> None:
        pass

    @abstractmethod
    def download(self, resource_type: Type, id: str, target: str) -> str:
        pass

    @abstractmethod
    def get_temporary_link(self, filename: str, target: str) -> str:
        pass



from typing import Type, List
from splight_abstract.client import AbstractClient


class AbstractRemoteClient(AbstractClient):
    def _parse_params(self, **kwargs):
        params = {}
        for key, value in kwargs.items():
            params[key] = value
            if isinstance(value, list):
                params[key] = ','.join(value)
        return params



from .base import SplightBaseModel
from typing import Optional


class Mapping(SplightBaseModel):
    id: Optional[str]
    name: Optional[str]
    description: Optional[str] = None
    asset_id: str
    attribute_id: str
    output_format: str



from pydantic import BaseModel
from requests import Session
from furl import furl
from typing import Type, List, Optional
from remote_splight_lib.auth.auth import SplightAuthToken
from splight_abstract.deployment.abstract import AbstractDeploymentClient
from remote_splight_lib.settings import settings
from splight_abstract.remote.abstract import AbstractRemoteClient
from splight_models.deployment import Deployment
from retry import retry
from requests.exceptions import (
    ConnectionError,
    Timeout
)
REQUEST_EXCEPTIONS = (ConnectionError, Timeout)


class DeploymentClient(AbstractDeploymentClient, AbstractRemoteClient):
    PATH = 'deployment'

    def __init__(self, namespace: str = "default", *args, **kwargs):
        super().__init__(namespace, *args, **kwargs)
        self._base_url = furl(settings.SPLIGHT_PLATFORM_API_HOST)
        token = SplightAuthToken(
            access_key=settings.SPLIGHT_ACCESS_ID,
            secret_key=settings.SPLIGHT_SECRET_KEY,
        )
        self._session = Session()
        self._session.headers.update(token.header)

    def save(self, instance: BaseModel) -> BaseModel:
        raise NotImplementedError

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=1)
    def _get(self, resource_type: Type, id: str = '', first=False, limit_: int = -1, skip_: int = 0, **kwargs) -> List[BaseModel]:

        if resource_type != Deployment:
            raise NotImplementedError
        response = self._list(
            first=first,
            limit_=limit_,
            skip_=skip_,
            **kwargs,
        )
        parsed = [
            resource_type.parse_obj(resource)
            for resource in response["results"]
        ]
        return parsed

    @retry(REQUEST_EXCEPTIONS, tries=3, delay=1)
    def _list(self, **kwargs):
        url = self._base_url / f"{self.PATH}/"
        params = self._parse_params(**kwargs)
        response = self._session.get(url, params=params)
        response.raise_for_status()
        return response.json()

    def delete(self, resource_type: Type, id: str) -> None:
        raise NotImplementedError

    def get_deployment_logs(self, id: str, limit: Optional[int] = None, since: Optional[str] = None, previous: bool = False) -> List[str]:
        raise NotImplementedError

    def get_capacity_options(self):
        raise NotImplementedError

    @classmethod
    def verify_header(cls, payload: bytes, signature: str) -> None:
        pass



import re
from typing import List, Any

class FilterMixin:
    reserved_filter_keys = ["__in", "__contains", "__lte", "__gte"]

    def _get_field_value(self, x: Any, field: str):
        for reserved_key in self.reserved_filter_keys:
            field = field.replace(reserved_key, "")

        if "__" in field:
            key, subkey = field.split("__")
            try:
                return x.get(key, {}).get(subkey)
            except:
                pass
            try:
                return getattr(x, key).get(subkey)
            except:
                return None
        return getattr(x, field)

    def _check_in(self, x: Any, field: str, value: Any):
        field = self._get_field_value(x,field)
        try:
            return field in value
        except:
            return False

    def _check_contains(self, x: Any, field: str, value: Any):
        field = self._get_field_value(x, field)
        try:
            return value in field
        except:
            return False

    def _filter(self, queryset: list, **kwargs) -> list:
        '''
        Filter a queryset by the given kwargs.
        '''
        field_filters = [
            lambda x, field=field, value=value: self._get_field_value(x, field) == value for field, value in kwargs.items() if "__" not in field
        ]
        in_filters = [
            lambda x, field=field, values=values: self._check_in(x, field, values) for field, values in kwargs.items() if field.endswith("__in")
        ]
        contains_filters = [
            lambda x, field=field, value=value: self._check_contains(x, field, value) for field, value in kwargs.items() if field.endswith("__contains")
        ]
        dict_filters = [
            lambda x, field=field, value=value: self._get_field_value(x, field) == value for field, value in kwargs.items() if "__" in field and not any(field.endswith(reserved_key) for reserved_key in self.reserved_filter_keys)
        ]
        filters = field_filters + in_filters + contains_filters + dict_filters

        return [obj for obj in queryset if all([f(obj) for f in filters])]

    def _validated_kwargs(self, allowed_fields: List[str], **kwargs):
        '''
        Validate the given kwargs.
        '''
        valid_kwargs = [
            f"{field}__in" for field in allowed_fields
        ] + [
            f"{field}__contains" for field in allowed_fields
        ] + [
            f"{field}__lte" for field in allowed_fields
        ] + [
            f"{field}__gte" for field in allowed_fields
        ] + allowed_fields

        invalid_kwargs = []
        for key in kwargs.keys():
            if key not in valid_kwargs and not any(key.startswith(f"{allowed_field}__") for allowed_field in allowed_fields):
                invalid_kwargs.append(key)

        for key in invalid_kwargs:
            kwargs.pop(key)
        return kwargs



import pysher
import requests
from furl import furl
from retry import retry
from typing import Callable, Dict
from splight_lib.logging import logging
from splight_abstract.communication import (
    AbstractCommunicationClient,
    ClientNotReady,
)
from splight_models.communication import CommunicationClientStatus, CommunicationContext, CommunicationEvent

from remote_splight_lib.auth.auth import SplightAuthToken
from remote_splight_lib.settings import settings
from remote_splight_lib.communication.classmap import CLASSMAP


logger = logging.getLogger()


class CommunicationFactory:

    def __init__(self, model):
        self._model = model

    def get_url(self):
        base_url = furl(settings.SPLIGHT_PLATFORM_API_HOST)
        return base_url / CLASSMAP.get(self._model)

    def get_headers(self):
        auth_token = SplightAuthToken(
            access_key=settings.SPLIGHT_ACCESS_ID,
            secret_key=settings.SPLIGHT_SECRET_KEY,
        )
        return auth_token.header

    def create(self, data: Dict):
        response = requests.post(self.get_url(), json=data, headers=self.get_headers())
        assert response.status_code == 201, f"Cant create communication {self._model}."
        return self._model.parse_obj(response.json())

    def get(self, params=None):
        response = requests.get(self.get_url(), params=params, headers=self.get_headers())
        assert response.status_code == 200, f"Cant fetch communication {self._model}."
        data = response.json()
        return self._model.parse_obj(data)


class CommunicationClient(AbstractCommunicationClient):

    def __init__(self, daemon: bool = True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._status = CommunicationClientStatus.STOPPED
        self._channel_bindings = []
        self._client, self._context = None, None
        self._system_bindings = [
            ("pusher:connection_established", self.__on_connection_established),
            ("pusher:connection_failed", self.__on_connection_failed),
            ("pusher:error", self.__on_error),
        ]
        self._daemon = daemon
        try:
            self.__load_context()
            self.__load_client()
            self.__check_readiness()
        except Exception as e:
            logger.exception(e)
            logger.warning("Failed to start communication client. Moving forward without remote commands.")
            self._status = CommunicationClientStatus.ERROR

    @property
    def status(self):
        return self._status

    @property
    def context(self):
        return self._context

    @retry(ClientNotReady, tries=3, delay=2, jitter=1)
    def __check_readiness(self):
        if self.status != CommunicationClientStatus.READY:
            raise ClientNotReady

    @retry(Exception, tries=3, delay=2, jitter=1)
    def __load_context(self):
        params = {"instance_id": self.instance_id}
        self._context = CommunicationFactory(CommunicationContext).get(params)

    @retry(Exception, tries=3, delay=2, jitter=1)
    def __load_client(self):
        if not self._context:
            raise NotImplementedError
        self._client = pysher.Pusher(
            key=self._context.key,
            auth_endpoint=self._context.auth_endpoint,
            auth_endpoint_headers=self._context.auth_headers,
            user_data=self._context.channel_data.dict() if self.context.channel_data else None,
            daemon=self._daemon,
        )
        self.__bind_system_events()
        self._client.connect()

    def __bind_system_events(self):
        if self._client is None:
            return
        for event, handler in self._system_bindings:
            self._client.connection.bind(event, handler)

    def __on_connection_established(self, data):
        self._channel = self._client.subscribe(self._context.channel)
        self._private_room_channel = self._client.subscribe(self._context.private_room_channel)
        for event_name, event_handler in self._channel_bindings:
            self._channel.bind(event_name, event_handler)
            self._private_room_channel.bind(event_name, event_handler)
        self._status = CommunicationClientStatus.READY

    def __on_connection_failed(self, data):
        self._status = CommunicationClientStatus.FAILED

    def __on_error(self, data):
        logger.error("Error on message", data)

    def bind(self, event_name: str, event_handler: Callable) -> None:
        self._channel_bindings.append((event_name, event_handler))
        if self.status != CommunicationClientStatus.READY:
            logger.warning("Bind events failed due to comm client is not ready")
            return
        self._channel.bind(event_name, event_handler)
        self._private_room_channel.bind(event_name, event_handler)

    def unbind(self, event_name: str, event_handler: Callable) -> None:
        # TODO implement this
        raise NotImplementedError

    def trigger(self, event: CommunicationEvent):
        return CommunicationFactory(CommunicationEvent).create(data=event.dict())

    def authenticate(self, channel_name: str, socket_id: str, custom_data: Dict = None) -> Dict:
        raise NotImplementedError
